/*
JRev: A JAVA MP3 Source Streamer
Copyright (C) 2000  Thomas Bohmbach, Jr.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package org.jrev;

import org.xml.sax.helpers.XMLReaderFactory;
import org.xml.sax.XMLReader;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;
import org.xml.sax.SAXParseException;
import org.xml.sax.SAXException;
import org.xml.sax.ContentHandler;
import org.xml.sax.ErrorHandler;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.apache.xerces.parsers.SAXParser;
import java.util.Hashtable;

/**
 * A class to handle the reading of an XML config file for the "out-of-the-box"
 * functionality of {@link JRev} and {@link FilePlaylist}.  After reading the
 * file, the {@link ConfigFile} will have constructed a <code>Hashtable</code>
 * of the various settings.  Getters can be called to get the various config
 * settings.  Note that getters are provided that return the appropriate types
 * (rather than generic <code>String</code>s).
 *
 * @author      pitboss
 * @version     1.0, 20000728
 * @since       1.0
 */
public class ConfigFile extends LoggerClient
{
    private static final String DRIVER_CLASS_NAME = "org.apache.xerces.parsers.SAXParser";

    //Tag names
    public static final String JREV_CONFIG_ELEMENT = "jrev_config";
    public static final String JREV_CONFIG_VERSION_ATTRIBUTE = "jrev_config_version";
    public static final String JREV_ELEMENT = "jrev";
    public static final String JREV_BUFFERSIZE_ATTRIBUTE = "jrev_buffersize";
    public static final String JREV_OVERHEAD_ATTRIBUTE = "jrev_overhead";
    public static final String JREV_PUBLIC_ATTRIBUTE = "jrev_public";
    public static final String JREV_ICY_ATTRIBUTE = "jrev_useicyheaders";
    public static final String JREV_BITRATE_ATTRIBUTE = "jrev_bitrate";
    public static final String JREV_LOG_ELEMENT = "jrev_log";
    public static final String JREV_QUIET_ATTRIBUTE = "jrev_quiet";
    public static final String JREV_DEBUG_ATTRIBUTE = "jrev_debug";
    public static final String JREV_NAME_ELEMENT = "jrev_name";
    public static final String JREV_DESCRIPTION_ELEMENT = "jrev_description";
    public static final String JREV_URL_ELEMENT = "jrev_url";
    public static final String JREV_GENRE_ELEMENT = "jrev_genre";
    public static final String JREV_SERVER_ELEMENT = "jrev_server";
    public static final String JREV_PORT_ATTRIBUTE = "jrev_port";
    public static final String JREV_PASSWORD_ATTRIBUTE = "jrev_password";
    public static final String JREV_MOUNTPOINT_ATTRIBUTE = "jrev_mountpoint";

    public static final String PLAYLIST_ELEMENT = "playlist";
    public static final String PLAYLIST_RANDOMIZE_ATTRIBUTE = "playlist_randomize";
    public static final String PLAYLIST_LOG_ELEMENT = "playlist_log";
    public static final String PLAYLIST_QUIET_ATTRIBUTE = "playlist_quiet";
    public static final String PLAYLIST_DEBUG_ATTRIBUTE = "playlist_debug";
    public static final String PLAYLIST_FILENAME_ELEMENT = "playlist_filename";
    public static final String PLAYLIST_HAS_HISTORY = "playlist_has_history";
    public static final String PLAYLIST_HISTORY_FILE_ELEMENT = "playlist_history_file";
    public static final String PLAYLIST_HISTORY_SIZE_ATTRIBUTE = "playlist_history_size";

    public static final String TRI_ELEMENT = "tri";
    public static final String TRI_LOG_ELEMENT = "tri_log";
    public static final String TRI_QUIET_ATTRIBUTE = "tri_quiet";
    public static final String TRI_DEBUG_ATTRIBUTE = "tri_debug";
    public static final String TRI_PORT_ELEMENT = "tri_port";
    public static final String TRI_PASSWORD_ELEMENT = "tri_password";

    private XMLReader itsParser = null;
    private Hashtable itsConfigHash = null;
    private boolean itsFinishedWithoutErrors = false;

    /**
     * An implementation of the SAX ContentHandler
     * (<code>org.xml.sax.ContentHandler</code>).  This class actually does
     * the work of adding configs to the internal <code>Hashtable</code>.  For
     * more info about the methods in this class, please see the JavaDoc for
     * <code>org.xml.sax.ContentHandler</code>.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     * @since       1.0
     */
    private class ConfigContentHandler implements ContentHandler
    {
        private ConfigFile itsConfigFile = null;
        private StringBuffer itsStringBuffer = new StringBuffer();

        public ConfigContentHandler()
        {
        }

        public void startDocument() throws SAXException
        {
            println("startDocument()", 4);
        }

        public void endDocument() throws SAXException
        {
            println("endDocument()", 4);
        }

        public void startElement(String namespaceURI,
    			                 String localName,
    			                 String qName,
    			                 Attributes atts)
        {
            println("startElement() : " +
                    namespaceURI + " : " +
                    localName + " : " +
                    qName, 4);

            String theAttributeValue = null;
            String theElementName = localName;

            itsStringBuffer.setLength(0);

            if (theElementName.equals(ConfigFile.JREV_CONFIG_ELEMENT))
            {
                theAttributeValue = atts.getValue(ConfigFile.JREV_CONFIG_VERSION_ATTRIBUTE);
                if (theAttributeValue != null)
                    setConfig(ConfigFile.JREV_CONFIG_VERSION_ATTRIBUTE, theAttributeValue);
                println(ConfigFile.JREV_CONFIG_VERSION_ATTRIBUTE + " = " + theAttributeValue, 4);
            }
            else if (theElementName.equals(ConfigFile.JREV_ELEMENT))
            {
                theAttributeValue = atts.getValue(ConfigFile.JREV_BUFFERSIZE_ATTRIBUTE);
                if (theAttributeValue != null)
                    setConfig(ConfigFile.JREV_BUFFERSIZE_ATTRIBUTE, theAttributeValue);
                println(ConfigFile.JREV_BUFFERSIZE_ATTRIBUTE + " = " + theAttributeValue, 4);

                theAttributeValue = atts.getValue(ConfigFile.JREV_OVERHEAD_ATTRIBUTE);
                if (theAttributeValue != null)
                    setConfig(ConfigFile.JREV_OVERHEAD_ATTRIBUTE, theAttributeValue);
                println(ConfigFile.JREV_OVERHEAD_ATTRIBUTE + " = " + theAttributeValue, 4);

                theAttributeValue = atts.getValue(ConfigFile.JREV_PUBLIC_ATTRIBUTE);
                if (theAttributeValue != null)
                    setConfig(ConfigFile.JREV_PUBLIC_ATTRIBUTE, theAttributeValue);
                println(ConfigFile.JREV_PUBLIC_ATTRIBUTE + " = " + theAttributeValue, 4);

                theAttributeValue = atts.getValue(ConfigFile.JREV_ICY_ATTRIBUTE);
                if (theAttributeValue != null)
                    setConfig(ConfigFile.JREV_ICY_ATTRIBUTE, theAttributeValue);
                println(ConfigFile.JREV_ICY_ATTRIBUTE + " = " + theAttributeValue, 4);

                theAttributeValue = atts.getValue(ConfigFile.JREV_BITRATE_ATTRIBUTE);
                if (theAttributeValue != null)
                    setConfig(ConfigFile.JREV_BITRATE_ATTRIBUTE, theAttributeValue);
                println(ConfigFile.JREV_BITRATE_ATTRIBUTE + " = " + theAttributeValue, 4);
            }
            else if (theElementName.equals(ConfigFile.JREV_LOG_ELEMENT))
            {
                theAttributeValue = atts.getValue(ConfigFile.JREV_QUIET_ATTRIBUTE);
                if (theAttributeValue != null)
                    setConfig(ConfigFile.JREV_QUIET_ATTRIBUTE, theAttributeValue);
                println(ConfigFile.JREV_QUIET_ATTRIBUTE + " = " + theAttributeValue, 4);

                theAttributeValue = atts.getValue(ConfigFile.JREV_DEBUG_ATTRIBUTE);
                if (theAttributeValue != null)
                    setConfig(ConfigFile.JREV_DEBUG_ATTRIBUTE, theAttributeValue);
                println(ConfigFile.JREV_DEBUG_ATTRIBUTE + " = " + theAttributeValue, 4);
            }
            else if (theElementName.equals(ConfigFile.JREV_SERVER_ELEMENT))
            {
                theAttributeValue = atts.getValue(ConfigFile.JREV_PORT_ATTRIBUTE);
                if (theAttributeValue != null)
                    setConfig(ConfigFile.JREV_PORT_ATTRIBUTE, theAttributeValue);
                println(ConfigFile.JREV_PORT_ATTRIBUTE + " = " + theAttributeValue, 4);

                theAttributeValue = atts.getValue(ConfigFile.JREV_PASSWORD_ATTRIBUTE);
                if (theAttributeValue != null)
                    setConfig(ConfigFile.JREV_PASSWORD_ATTRIBUTE, theAttributeValue);
                println(ConfigFile.JREV_PASSWORD_ATTRIBUTE + " = " + theAttributeValue, 4);

                theAttributeValue = atts.getValue(ConfigFile.JREV_MOUNTPOINT_ATTRIBUTE);
                if (theAttributeValue != null)
                    setConfig(ConfigFile.JREV_MOUNTPOINT_ATTRIBUTE, theAttributeValue);
                println(ConfigFile.JREV_MOUNTPOINT_ATTRIBUTE + " = " + theAttributeValue, 4);
            }
            else if (theElementName.equals(ConfigFile.PLAYLIST_ELEMENT))
            {
                theAttributeValue = atts.getValue(ConfigFile.PLAYLIST_RANDOMIZE_ATTRIBUTE);
                if (theAttributeValue != null)
                    setConfig(ConfigFile.PLAYLIST_RANDOMIZE_ATTRIBUTE, theAttributeValue);
                println(ConfigFile.PLAYLIST_RANDOMIZE_ATTRIBUTE + " = " + theAttributeValue, 4);
            }
            else if (theElementName.equals(ConfigFile.PLAYLIST_LOG_ELEMENT))
            {
                theAttributeValue = atts.getValue(ConfigFile.PLAYLIST_QUIET_ATTRIBUTE);
                if (theAttributeValue != null)
                    setConfig(ConfigFile.PLAYLIST_QUIET_ATTRIBUTE, theAttributeValue);
                println(ConfigFile.PLAYLIST_QUIET_ATTRIBUTE + " = " + theAttributeValue, 4);

                theAttributeValue = atts.getValue(ConfigFile.PLAYLIST_DEBUG_ATTRIBUTE);
                if (theAttributeValue != null)
                    setConfig(ConfigFile.PLAYLIST_DEBUG_ATTRIBUTE, theAttributeValue);
                println(ConfigFile.PLAYLIST_DEBUG_ATTRIBUTE + " = " + theAttributeValue, 4);
            }
            else if (theElementName.equals(ConfigFile.PLAYLIST_HISTORY_FILE_ELEMENT))
            {
                theAttributeValue = atts.getValue(ConfigFile.PLAYLIST_HISTORY_SIZE_ATTRIBUTE);
                if (theAttributeValue != null)
                    setConfig(ConfigFile.PLAYLIST_HISTORY_SIZE_ATTRIBUTE, theAttributeValue);
                println(ConfigFile.PLAYLIST_HISTORY_SIZE_ATTRIBUTE + " = " + theAttributeValue, 4);

                setConfig(ConfigFile.PLAYLIST_HAS_HISTORY, "true");
                println(ConfigFile.PLAYLIST_HAS_HISTORY + " = true", 4);
            }
            else if (theElementName.equals(ConfigFile.TRI_ELEMENT))
            {
                setConfig(ConfigFile.TRI_ELEMENT, "true");
                println(ConfigFile.TRI_ELEMENT + " = true", 4);
            }
            else if (theElementName.equals(ConfigFile.TRI_LOG_ELEMENT))
            {
                theAttributeValue = atts.getValue(ConfigFile.TRI_QUIET_ATTRIBUTE);
                if (theAttributeValue != null)
                    setConfig(ConfigFile.TRI_QUIET_ATTRIBUTE, theAttributeValue);
                println(ConfigFile.TRI_QUIET_ATTRIBUTE + " = " + theAttributeValue, 4);

                theAttributeValue = atts.getValue(ConfigFile.TRI_DEBUG_ATTRIBUTE);
                if (theAttributeValue != null)
                    setConfig(ConfigFile.TRI_DEBUG_ATTRIBUTE, theAttributeValue);
                println(ConfigFile.TRI_DEBUG_ATTRIBUTE + " = " + theAttributeValue, 4);
            }
        }

        public void endElement(String namespaceURI,
    		                   String localName,
    			               String qName)
        {
            println("endElement() : " +
                    namespaceURI + " : " +
                    localName + " : " +
                    qName, 4);

        	String theElementName = localName;
		    String theString = null;

            if (theElementName.equals(JREV_LOG_ELEMENT) ||
                theElementName.equals(JREV_NAME_ELEMENT) ||
                theElementName.equals(JREV_DESCRIPTION_ELEMENT) ||
                theElementName.equals(JREV_URL_ELEMENT) ||
                theElementName.equals(JREV_GENRE_ELEMENT) ||
                theElementName.equals(JREV_SERVER_ELEMENT) ||
                theElementName.equals(PLAYLIST_LOG_ELEMENT) ||
                theElementName.equals(PLAYLIST_FILENAME_ELEMENT) ||
                theElementName.equals(PLAYLIST_HISTORY_FILE_ELEMENT) ||
                theElementName.equals(TRI_LOG_ELEMENT) ||
                theElementName.equals(TRI_PORT_ELEMENT) ||
                theElementName.equals(TRI_PASSWORD_ELEMENT))
            {
                theString = itsStringBuffer.toString().trim();
                if ((theString != null) && !(theString.equals("")))
                    setConfig(theElementName, theString);
                println(theElementName + " = " + theString, 4);
            }
        }

        public void characters(char[] ch, int start, int length)
            throws SAXException
        {
            itsStringBuffer.append(ch, start, length);
        }

        public void ignorableWhitespace(char buf [], int offset, int len)
            throws SAXException
        {
        }

        public void processingInstruction (String target, String data)
            throws SAXException
        {
        }

        public void startPrefixMapping(String prefix,
    			                       String uri)
        {
        }

        public void endPrefixMapping(String prefix)
        {
        }

        public void skippedEntity(String name)
        {
        }

        public void setDocumentLocator(Locator locator)
        {
        }
    }

    /**
     * An implementation of the SAX ErrorHandler
     * (<code>org.xml.sax.ErrorHandler</code>).  This class is needed to
     * validate the XML config file.  For more info about the methods in this
     * class, please see the JavaDoc for
     * <code>org.xml.sax.ErrorHandler</code>.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     * @since       1.0
     */
    private class ConfigErrorHandler implements ErrorHandler
    {
        public ConfigErrorHandler()
        {
        }

        public void warning(SAXParseException e)
            throws SAXException
        {
        }

        public void error(SAXParseException e)
            throws SAXException
        {
            throw e;
        }
        public void fatalError(SAXParseException e)
            throws SAXException
        {
            throw e;
        }
    }

    /**
     * Creates an instance of a <code>ConfigFile</code> that reads its configs
     * from the XML file with the given file name.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inFilename    the file name of the XML config file
     */
    public ConfigFile(String inFilename)
    {
	    try
        {
            //Make sure we can print to the screen
            Logger theLogger = new Logger(false, null);
            setLogger(theLogger);

            //Set up our hashtable
            itsConfigHash = new Hashtable();

	        itsParser = XMLReaderFactory.createXMLReader(DRIVER_CLASS_NAME);

            itsParser.setFeature("http://xml.org/sax/features/validation", true);

            itsParser.setContentHandler(new ConfigContentHandler());
            itsParser.setErrorHandler(new ConfigErrorHandler());

            println("Reading config file (" + inFilename + ")...");
            itsParser.parse(inFilename);
            itsFinishedWithoutErrors = true;
        }
        catch (SAXNotRecognizedException snre)
        {
	        println(snre.getMessage());
        }
        catch (SAXNotSupportedException snse)
        {
    	    println(snse.getMessage());
        }
        catch (SAXParseException spe)
        {
            println("XML Parse Exception in " +
                    spe.getSystemId() +
                    " at line " + spe.getLineNumber() +
                    " : " + spe.getMessage());

	        Exception  x = spe;
	        if (spe.getException() != null)
            {
		        x = spe.getException();
		        println(x);
	        }
        }
        catch (SAXException sxe)
        {
	        println(sxe.getMessage());

	        Exception  x = sxe;
	        if (sxe.getException() != null)
            {
		        x = sxe.getException();
		        println( x );
	        }
        }
        catch (Exception e)
        {
            println(e);
        }
    }

    /**
     * Returns whether the <code>ConfigFile</code> was able to finish reading
     * without any errors.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      whether this is a complete <code>ConfigFile</code>
     */
    public boolean complete()
    {
        return itsFinishedWithoutErrors;
    }

    /**
     * Sets the given config to the given <code>String</code>.  If the value is
     * <code>null</code> it will set the value to the empty string "".
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inConfigName      the config <code>String</code> to set
     * @param inConfigValue     the <code>String</code> value
     */
    private void setConfig(String inConfigName, String inConfigValue)
    {
        String theValue = "";

        if (inConfigName != null)
        {
            if (inConfigValue != null)
                theValue = inConfigValue;

            itsConfigHash.put(inConfigName, theValue);
        }
    }

    /**
     * Gets the value for the given config.  If the value is the empty string,
     * <code>null</code> will be returned.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inConfigName      the config <code>String</code> to get
     *
     * @return                  the <code>String</code> value; <code>null</code>
     *                          if the value is the empty string
     */
    private String getConfig(String inConfigName)
    {
        String theValue = null;

        if (inConfigName != null)
        {
            theValue = (String) itsConfigHash.get(inConfigName);

            if ((theValue != null) && (theValue.equals("")))
                theValue = null;
        }

        return theValue;
    }

    /**
     * Gets the boolean value of a string with a given default if the string
     * is null.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inString      the <code>String</code> to convert to a boolean
     * @param inDefault     the default value
     *
     * @return              the boolean value
     */
    private boolean stringToBoolean(String inString, boolean inDefault)
    {
        boolean theBoolean = inDefault;

        if (inString != null)
        {
            if (inString.equalsIgnoreCase("true") ||
                inString.equalsIgnoreCase("yes") ||
                inString.equals("1"))
            {
                theBoolean = true;
            }
            else if (inString.equalsIgnoreCase("false") ||
                     inString.equalsIgnoreCase("no") ||
                     inString.equals("0"))
            {
                theBoolean = false;
            }
        }

        return theBoolean;
    }

    /**
     * Gets the int value of a string with a given default if the string
     * is null.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inString      the <code>String</code> to convert to an int
     * @param inDefault     the default value
     *
     * @return              the int value
     */
    private int stringToInt(String inString, int inDefault)
    {
        int theInt = inDefault;

        if (inString != null)
            theInt = Integer.parseInt(inString);

        return theInt;
    }

    /**
     * Gets the <code>String</code> value of a string with a given default if
     * the string is null.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inString      the <code>String</code> to convert to a
     *                      <code>String</code>
     * @param inDefault     the default value
     *
     * @return              the <code>String</code> value
     */
    private String stringToString(String inString, String inDefault)
    {
        String theString = inDefault;

        if (inString != null)
            theString = inString;

        return theString;
    }

    /**
     * Returns the <code>JRev</code> buffersize config (Default: 4000).
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>JRev</code> buffersize config
     */
    public int getBuffersize()
    {
        String theStringValue = getConfig(JREV_BUFFERSIZE_ATTRIBUTE);

        return (stringToInt(theStringValue, 4000));
    }

    /**
     * Returns the <code>JRev</code> overhead config (Default: 5).
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>JRev</code> overhead config
     */
    public int getOverhead()
    {
        String theStringValue = getConfig(JREV_OVERHEAD_ATTRIBUTE);

        return (stringToInt(theStringValue, 5));
    }

    /**
     * Returns the <code>JRev</code> public config (Default: false).
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>JRev</code> public config
     */
    public boolean getIsPublic()
    {
        String theStringValue = getConfig(JREV_PUBLIC_ATTRIBUTE);

        return (stringToBoolean(theStringValue, false));
    }

    /**
     * Returns the <code>JRev</code> ICY headers config (Default: true).
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>JRev</code> ICY headers config
     */
    public boolean getIsIcyHeaders()
    {
        String theStringValue = getConfig(JREV_ICY_ATTRIBUTE);

        return (stringToBoolean(theStringValue, true));
    }

    /**
     * Returns the <code>JRev</code> bitrate config (Default: 128).
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>JRev</code> bitrate config
     */
    public int getBitrate()
    {
        String theStringValue = getConfig(JREV_BITRATE_ATTRIBUTE);

        return (stringToInt(theStringValue, 128));
    }

    /**
     * Returns the <code>JRev</code> log file name config
     * (Default: <code>null</code>).
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>JRev</code> log file name config
     */
    public String getJRevLogFilename()
    {
        String theStringValue = getConfig(JREV_LOG_ELEMENT);

        return (stringToString(theStringValue, null));
    }

    /**
     * Returns the <code>JRev</code> quiet config (Default: false).
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>JRev</code> quiet config
     */
    public boolean getIsJRevQuiet()
    {
        String theStringValue = getConfig(JREV_QUIET_ATTRIBUTE);

        return (stringToBoolean(theStringValue, false));
    }

    /**
     * Returns the <code>JRev</code> debug level config (Default: 0).
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>JRev</code> debug level config
     */
    public int getJRevOutputLevel()
    {
        String theStringValue = getConfig(JREV_DEBUG_ATTRIBUTE);

        return (stringToInt(theStringValue, 0));
    }

    /**
     * Returns the <code>JRev</code> name config (Default: "NoName").
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>JRev</code> name config
     */
    public String getName()
    {
        String theStringValue = getConfig(JREV_NAME_ELEMENT);

        return (stringToString(theStringValue, "NoName"));
    }

    /**
     * Returns the <code>JRev</code> description config
     * (Default: "NoDescription").
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>JRev</code> description config
     */
    public String getDescription()
    {
        String theStringValue = getConfig(JREV_DESCRIPTION_ELEMENT);

        return (stringToString(theStringValue, "NoDescription"));
    }

    /**
     * Returns the <code>JRev</code> URL config
     * (Default: "http://www.no.url.com").
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>JRev</code> URL config
     */
    public String getURL()
    {
        String theStringValue = getConfig(JREV_URL_ELEMENT);

        return (stringToString(theStringValue, "http://www.no.url.com"));
    }

    /**
     * Returns the <code>JRev</code> genre config (Default: "NoGenre").
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>JRev</code> genre config
     */
    public String getGenre()
    {
        String theStringValue = getConfig(JREV_GENRE_ELEMENT);

        return (stringToString(theStringValue, "NoGenre"));
    }

    /**
     * Returns the <code>JRev</code> server config (Default: "127.0.0.1").
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>JRev</code> server config
     */
    public String getServer()
    {
        String theStringValue = getConfig(JREV_SERVER_ELEMENT);

        return (stringToString(theStringValue, "127.0.0.1"));
    }

    /**
     * Returns the <code>JRev</code> port config (Default: 8000).
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>JRev</code> port config
     */
    public int getJRevPort()
    {
        String theStringValue = getConfig(JREV_PORT_ATTRIBUTE);

        return (stringToInt(theStringValue, 8000));
    }

    /**
     * Returns the <code>JRev</code> password config (Default: "letmein").
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>JRev</code> password config
     */
    public String getJRevPassword()
    {
        String theStringValue = getConfig(JREV_PASSWORD_ATTRIBUTE);

        return (stringToString(theStringValue, "letmein"));
    }

    /**
     * Returns the <code>JRev</code> mountpoint config
     * (Default: <code>null</code>).
     *
     * @author      Bryan Buchanan
     * @version     1.2, 20000822
     *
     * @return      the <code>JRev</code> mountpoint config
     */
    public String getJRevMountPoint()
    {
        String theStringValue = getConfig(JREV_MOUNTPOINT_ATTRIBUTE);

        return (stringToString(theStringValue, null));
    }

    /**
     * Returns the <code>FilePlaylist</code> randomize config (Default: false).
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>FilePlaylist</code> randomize config
     */
    public boolean getIsRandomizePlaylist()
    {
        String theStringValue = getConfig(PLAYLIST_RANDOMIZE_ATTRIBUTE);

        return (stringToBoolean(theStringValue, false));
    }

    /**
     * Returns the <code>FilePlaylist</code> log file name config
     * (Default: <code>null</code>).
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>FilePlaylist</code> log file name config
     */
    public String getPlaylistLogFilename()
    {
        String theStringValue = getConfig(PLAYLIST_LOG_ELEMENT);

        return (stringToString(theStringValue, null));
    }

    /**
     * Returns the <code>FilePlaylist</code> quiet config (Default: true).
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>FilePlaylist</code> quiet config
     */
    public boolean getIsPlaylistQuiet()
    {
        String theStringValue = getConfig(PLAYLIST_QUIET_ATTRIBUTE);

        return (stringToBoolean(theStringValue, true));
    }

    /**
     * Returns the <code>FilePlaylist</code> debug level config (Default: 0).
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>FilePlaylist</code> debug level config
     */
    public int getPlaylistOutputLevel()
    {
        String theStringValue = getConfig(PLAYLIST_DEBUG_ATTRIBUTE);

        return (stringToInt(theStringValue, 0));
    }

    /**
     * Returns the <code>FilePlaylist</code> file name config
     * (Default: "mp3playlist.m3u").
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>FilePlaylist</code> file name config
     */
    public String getPlaylistFilename()
    {
        String theStringValue = getConfig(PLAYLIST_FILENAME_ELEMENT);

        return (stringToString(theStringValue, "mp3playlist.m3u"));
    }

    /**
     * Returns the has <code>FilePlaylist</code> history file config
     * (Default: false).
     *
     * @author      pitboss
     * @version     1.2, 20000823
     *
     * @return      the has <code>FilePlaylist</code> history file config
     */
    public boolean getHasFilePlaylistHistory()
    {
        String theStringValue = getConfig(PLAYLIST_HAS_HISTORY);

        return (stringToBoolean(theStringValue, false));
    }

    /**
     * Returns the <code>FilePlaylist</code> history file name config
     * (Default: "history.xml").
     *
     * @author      pitboss
     * @version     1.2, 20000823
     *
     * @return      the <code>FilePlaylist</code> history file name config
     */
    public String getPlaylistHistoryFilename()
    {
        String theStringValue = getConfig(PLAYLIST_HISTORY_FILE_ELEMENT);

        return (stringToString(theStringValue, "history.xml"));
    }

    /**
     * Returns the <code>FilePlaylist</code> history size config (Default: 10).
     *
     * @author      pitboss
     * @version     1.2, 20000823
     *
     * @return      the <code>FilePlaylist</code> history size config
     */
    public int getPlaylistHistorySize()
    {
        String theStringValue = getConfig(PLAYLIST_HISTORY_SIZE_ATTRIBUTE);

        return (stringToInt(theStringValue, 10));
    }

    /**
     * Returns the has <code>TelnetRemoteInterface</code> config
     * (Default: false).
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the has <code>TelnetRemoteInterface</code> config
     */
    public boolean getHasTRI()
    {
        String theStringValue = getConfig(TRI_ELEMENT);

        return (stringToBoolean(theStringValue, false));
    }

    /**
     * Returns the <code>TelnetRemoteInterface</code> log file name config
     * (Default: <code>null</code>).
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>TelnetRemoteInterface</code> log file name config
     */
    public String getTRILogFilename()
    {
        String theStringValue = getConfig(TRI_LOG_ELEMENT);

        return (stringToString(theStringValue, null));
    }

    /**
     * Returns the <code>TelnetRemoteInterface</code> quiet config (Default: true).
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>TelnetRemoteInterface</code> quiet config
     */
    public boolean getIsTRIQuiet()
    {
        String theStringValue = getConfig(TRI_QUIET_ATTRIBUTE);

        return (stringToBoolean(theStringValue, true));
    }

    /**
     * Returns the <code>TelnetRemoteInterface</code> debug level config
     * (Default: 0).
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>TelnetRemoteInterface</code> debug level config
     */
    public int getTRIOutputLevel()
    {
        String theStringValue = getConfig(TRI_DEBUG_ATTRIBUTE);

        return (stringToInt(theStringValue, 0));
    }

    /**
     * Returns the <code>TelnetRemoteInterface</code> port config
     * (Default: 8002).
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>TelnetRemoteInterface</code> port config
     */
    public int getTRIPort()
    {
        String theStringValue = getConfig(TRI_PORT_ELEMENT);

        return (stringToInt(theStringValue, 8002));
    }

    /**
     * Returns the <code>TelnetRemoteInterface</code> password config
     * (Default: "letmein").
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>TelnetRemoteInterface</code> password config
     */
    public String getTRIPassword()
    {
        String theStringValue = getConfig(TRI_PASSWORD_ELEMENT);

        return (stringToString(theStringValue, "letmein"));
    }
}
