/*
JRev: A JAVA MP3 Source Streamer
Copyright (C) 2000  Thomas Bohmbach, Jr.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package org.jrev;

import java.util.Vector;
import java.util.Random;
import java.lang.InterruptedException;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.PrintWriter;

/**
 * An implementation of {@link Playlist} to read in a list of MP3 files and
 * provide them as {@link MP3FileSource}s.  The class supports logging of source
 * progress to a {@link Logger} as well as the ability to randomize the initial
 * list.  When a <code>FilePlaylist</code> has exhausted its list of
 * <code>MP3FileSource</code>s, its will wrap and start providing sources from
 * the beginning of the list.  If it was set to randomize the list, it will do
 * so again before wrapping.
 *
 * @author      pitboss
 * @version     1.0, 20000728
 * @since       1.0
 * @see         Playlist
 * @see         MP3FileSource
 * @see         Logger
 */
public class FilePlaylist extends LoggerClient implements Playlist
{
    /**
     * An internal private thread class to monitor and output the progress of
     * this {@link FilePlaylist}'s {@link MP3FileSource}s.  Since this is
     * a private inner class, it can (and does) use its parent
     * <code>FilePlaylist</code>'s ouput methods (which it in turn inherits from
     * {@link LoggerClient}).
     *
     * @author      pitboss
     * @version     1.0, 20000728
     * @since       1.0
     * @see         FilePlaylist
     * @see         MP3FileSource
     * @see         LoggerClient
     */
    private class FilePlaylistLogThread extends Thread
    {
        //The total length of the progress "bar" (in characters) to output
        private int PROGRESS_BAR_LENGTH = 76;

        private FilePlaylist itsPlaylist = null;
        private boolean itsIsStop = false;

        /**
         * Creates a <code>FilePlaylistLogThread</code> that will monitor and
         * the output the progress of the passed in <code>FilePlaylist</code>.
         *
         * @author      pitboss
         * @version     1.0, 20000728
         *
         * @param inPlaylist    the <code>FilePlaylist</code> to monitor
         */
        public FilePlaylistLogThread(FilePlaylist inPlaylist)
        {
            itsPlaylist = inPlaylist;
        }

        /**
         * Stops this thread from monitoring and logging progress.
         *
         * @author      pitboss
         * @version     1.0, 20000728
         */
        public void stopLogging()
        {
            itsIsStop = true;
        }

        /**
         * A utility method for turning seconds into MM:SS
         *
         * @author      pitboss
         * @version     1.1, 20000807
         *
         * @param inSeconds     the number of seconds to convert
         *
         * @return              A string in the form of MM:SS
         */
        private String secondsToTimeString(long inSeconds)
        {
            long theMinutes = inSeconds / 60;
            long theSeconds = inSeconds - (theMinutes * 60);
            StringBuffer theMMSSStringBuffer = new StringBuffer();

            theMMSSStringBuffer.append(theMinutes);

            theMMSSStringBuffer.append(":");

            if (theSeconds < 10)
                theMMSSStringBuffer.append("0");
            theMMSSStringBuffer.append(theSeconds);

            return (theMMSSStringBuffer.toString());
        }

        /**
         * Returns the info about a source in a formatted string.
         *
         * @author      pitboss
         * @version     1.1, 20000807
         *
         * @param inSource      the <code>MP3FileSource</code> to get the info
         *                      string for
         *
         * @return              the formatted info string, null if
         *                      <code>inSource</code> is null
         */
        private String getSourceInfoString(MP3FileSource inSource)
        {
            String theString = null;
            StringBuffer theStringBuffer = new StringBuffer();
            int theInt = 0;

            if (inSource != null)
            {
                //Append the filename
                theString = inSource.name();
                if (theString != null)
                    theStringBuffer.append("Filename:   " + theString + "\n");

                //Append the title
                theString = inSource.title();
                if (theString != null)
                    theStringBuffer.append("Title:      " + theString + "\n");

                //Append the artist
                theString = inSource.artist();
                if (theString != null)
                    theStringBuffer.append("Artist:     " + theString + "\n");

                //Append the album
                theString = inSource.album();
                if (theString != null)
                    theStringBuffer.append("Album:      " + theString + "\n");

                //Append the time
                long theTimeInSeconds = inSource.lengthInSeconds();
                if (theTimeInSeconds != 0)
                {
                    theString = secondsToTimeString(theTimeInSeconds);
                    theStringBuffer.append("Time:       " + theString + "\n");
                }

                theString = theStringBuffer.toString();
            }

            return theString;
        }

        /**
         * The main thread method.  It constantly loops, checking the progress
         * of the current track and outputting that information to its
         * <code>FilePlaylist</code>'s <code>Logger</code>.  Since it is only
         * checking every second, it assumes that if the source has changed
         * since the last time it checked that the previous source completed
         * normally.  This means that if the <code>JRev</code> engine skipped
         * the previous song part way through, the progress bar will still be
         * completed by this logger thread.
         *
         * @author      pitboss
         * @version     1.0, 20000728
         */
        public void run()
        {
            MP3FileSource theCurrentSource = null;
            MP3FileSource theNewSource = null;
            boolean theHasPrintedTitle = false;
            int theProgress = 0;
            int theNewProgress = 0;
            float thePercentComplete = 0;
            String theString = null;

            while ((itsPlaylist != null) && !itsIsStop)
            {
                theNewSource = (MP3FileSource) itsPlaylist.currentSource();

                //Is this a new source?
                if (theNewSource != theCurrentSource)
                {
                    if (theCurrentSource != null)
                    {
                        //Finish the progress bar
                        theNewProgress = PROGRESS_BAR_LENGTH;

                        for (int i = 0; i < theNewProgress - theProgress; i++)
                            print("-");
                        print(">");
                        println();
                    }

                    theCurrentSource = theNewSource;

                    //Reset our initial progress
                    theProgress = 0;
                    theNewProgress = 0;
                    theHasPrintedTitle = false;
                }
                else //This is the same source as the last time we checked
                {
                    if (theCurrentSource != null)
                    {
                        thePercentComplete = theCurrentSource.percentComplete();
                        theNewProgress =
                            (int)(thePercentComplete *
                                  (float)PROGRESS_BAR_LENGTH);

                        //Have we just started this source?
                        if ((theProgress == 0) &&
                            (thePercentComplete > 0.0F) &&
                            !theHasPrintedTitle)
                        {
                            //Print a newline to divide sources
                            print("\n");

                            //Print the index
                            println("Streaming Source #" +
                                    itsPlaylist.currentIndex() +
                                    ":");

                            //Print the song info
                            theString = getSourceInfoString(theNewSource);
                            if (theString != null)
                                print(theString);

                            //Start the next progress bar
                            print("<");

                            theHasPrintedTitle = true;
                        }

                        //Print our progress since the last time we checked
                        for (int i = 0; i < theNewProgress - theProgress; i++)
                            print("-");

                        theProgress = theNewProgress;
                    }
                }

                //Sleep
                try
                {
                    sleep(1000);
                }
                catch (InterruptedException ie)
                {
                    println(ie);
                }
            }
        }
    }

    /**
     * An internal private source history class.
     *
     * @author      pitboss
     * @version     1.2, 20000823
     * @since       1.2
     * @see         MP3FileSource
     */
    private class FilePlaylistHistoryEntry
    {
        public MP3FileSource itsSource = null;
        public String itsDateString = null;

        public FilePlaylistHistoryEntry(MP3FileSource inSource,
                                        String inDateString)
        {
            itsSource = inSource;
            itsDateString = inDateString;
        }
    }

    //FilePlaylist declarations begin here
    //Tag names
    public static final String FILE_PLAYLIST_HISTORY_ELEMENT =
        "file_playlist_history";
    public static final String FILE_PLAYLIST_HISTORY_DTD =
        "file_playlist_history.dtd";
    public static final String MP3_FILE_SOURCE_ELEMENT = "mp3_file_source";
    public static final String TIME_ATTRIBUTE = "mp3_start_time";
    public static final String FILENAME_ELEMENT = "mp3_filename";
    public static final String TITLE_ELEMENT = "mp3_title";
    public static final String ARTIST_ELEMENT = "mp3_artist";
    public static final String ALBUM_ELEMENT = "mp3_album";
    public static final String LENGTH_IN_SECONDS_ELEMENT =
        "mp3_length_in_seconds";


    private Vector itsSourceVector = null;
    private int itsIndex = -1;
    private String itsFilename = null;
    private boolean itsIsRandomize = false;
    private String itsXMLHistoryFilename = null;
    private int itsXMLHistorySize = 0;
    private Vector itsHistoryVector = null;
    private FilePlaylistLogThread itsLoggingThread = null;
    private StreamSource itsSource = null;

    /**
     * Creates a <code>FilePlaylist</code> that reads in and creates an inital
     * list of <code>MP3FileSource</code>s from the given playlist file.  No
     * logging of playlist information will be done.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inFilename    the file to read its <code>MP3FileSource</code>
     *                      file names from
     */
    public FilePlaylist(String inFilename)
    {
        itsFilename = inFilename;
        init();
    }

    /**
     * Creates a <code>FilePlaylist</code> that reads in and creates an inital
     * list of <code>MP3FileSource</code>s from the given playlist file.  It
     * will also set its <code>Logger</code> to the given <code>Logger</code>.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inFilename    the file name to read its <code>MP3FileSource</code>
     *                      file names from
     * @param inLogger      the <code>Logger</code> to send output to
     */
    public FilePlaylist(String inFilename, Logger inLogger)
    {
        itsFilename = inFilename;
        setLogger(inLogger);
        init();
    }

    /**
     * Creates a <code>FilePlaylist</code> that reads in and creates an inital
     * list of <code>MP3FileSource</code>s from the given playlist file.
     * Additionally, it also allows this list to be randomized once read into
     * memory.  No logging of playlist information will be done.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inFilename        the file to read its <code>MP3FileSource</code>
     *                          file names from
     * @param inIsRandomize     whether to randomize the sources
     */
    public FilePlaylist(String inFilename, boolean inIsRandomize)
    {
        itsFilename = inFilename;
        itsIsRandomize = inIsRandomize;
        init();
    }

    /**
     * Creates a <code>FilePlaylist</code> that reads in and creates an inital
     * list of <code>MP3FileSource</code>s from the given playlist file.
     * Additionally, it also allows this list to be randomized once read into
     * memory.  It will also set its <code>Logger</code> to the given
     * <code>Logger</code>.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inFilename        the file to read its <code>MP3FileSource</code>
     *                          file names from
     * @param inIsRandomize     whether to randomize the sources
     * @param inLogger          the <code>Logger</code> to send output to
     */
    public FilePlaylist(String inFilename,
                        boolean inIsRandomize,
                        Logger inLogger)
    {
        itsFilename = inFilename;
        itsIsRandomize = inIsRandomize;
        setLogger(inLogger);
        init();
    }

    /**
     * Creates a <code>FilePlaylist</code> that reads in and creates an inital
     * list of <code>MP3FileSource</code>s from the given playlist file.
     * Additionally, it also allows this list to be randomized once read into
     * memory, sets its <code>Logger</code> to the given <code>Logger</code>,
     * and sets its XML source history filename.
     *
     * @author      pitboss
     * @version     1.2, 20000822
     *
     * @param inFilename            the file to read its
     *                              <code>MP3FileSource</code> file names from
     * @param inIsRandomize         whether to randomize the sources
     * @param inLogger              the <code>Logger</code> to send output to
     * @param inXMLHistoryFilename  the file to output the XML source history to
     * @param inXMLHistorySize      the number of songs to keep in the XML
     *                              source history
     */
    public FilePlaylist(String inFilename,
                        boolean inIsRandomize,
                        Logger inLogger,
                        String inXMLHistoryFilename,
                        int inXMLHistorySize)
    {
        itsFilename = inFilename;
        itsIsRandomize = inIsRandomize;
        itsXMLHistoryFilename = inXMLHistoryFilename;
        itsXMLHistorySize = inXMLHistorySize;
        setLogger(inLogger);
        init();
    }

    /**
     * A generic initialization method for all constructors.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    private void init()
    {
        //Try to read the playlist
        itsSourceVector = readPlaylist();

        if (itsSourceVector != null)
        {
            //Do we need a history vector?
            if ((itsXMLHistoryFilename != null) && (itsXMLHistorySize > 0))
                itsHistoryVector = new Vector();

            //Do we randomize the list?
            if (itsIsRandomize)
                randomize();

            //Set the index to the first source
            setIndexAndSource(0);
        }
    }

    /**
     * Cleans up this <code>FilePlaylist</code> upon garbage collection.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    protected void finalize()
    {
        close();
    }

    /**
     * Cleans up this <code>FilePlaylist</code>.  Basically it just stops its
     * <code>FilePlaylistLogThread</code>.
     *
     * @author      pitboss
     * @version     1.1, 20000813
     */
    public void close()
    {
        stopLogging();
    }

    /**
     * Stops its <code>FilePlaylistLogThread</code> and all further output.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    public void stopLogging()
    {
        println("Stopping logging thread...");
        setLogger(null);
    }

    /**
     * Sets the <code>Logger</code> to send output to.  Also starts a
     * <code>FilePlaylistLogThread</code> to monitor source progress.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inLogger      the <code>Logger</code> to send output to
     */
    public void setLogger(Logger inLogger)
    {
        if (itsLoggingThread != null)
            itsLoggingThread.stopLogging();

        super.setLogger(inLogger);

        if (getLogger() != null)
        {
            println("Starting logging thread...");
            itsLoggingThread = new FilePlaylistLogThread(this);
            itsLoggingThread.start();
        }
    }

    /**
     * Returns the total number of sources in its source vector.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the number of sources
     */
    public int length()
    {
        int theLength = 0;

        if (itsSourceVector != null)
            theLength = itsSourceVector.size();

        return theLength;
    }

    /**
     * Returns its current index.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the current index
     */
    public int currentIndex()
    {
        return itsIndex;
    }

    /**
     * Returns the index of its current source.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the index of the current source, <code>-1</code> if the
     *              current source is <code>null</code> or the source vector
     *              is <code>null</code>
     */
    public int currentSourceIndex()
    {
        int theIndex = -1;

        if ((itsSourceVector != null) && (itsSource != null))
            theIndex = itsSourceVector.indexOf(itsSource);

        return theIndex;
    }

    /**
     * Sets its current index.  Note that this method doesn't actually change
     * the current source.  It only changes the current index.  If the given
     * value is invalid (less than -1 or greater than or equal to its length)
     * the current index will not be set.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inIndex      the new value for the current index
     */
    public void setIndex(int inIndex)
    {
        if ((inIndex >= -1) && (inIndex < length()))
            itsIndex = inIndex;
    }

    /**
     * Sets both its current index and its current source.  Its current source
     * will be set to the new current index entry in its source vector.
     * Wrapping calculations are done to determine if the playlist needs to be
     * started over.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inIndex      the new value for the current index
     */
    private void setIndexAndSource(int inIndex)
    {
        int theNewIndex = inIndex;
        int theLength = length();

        if (theNewIndex < 0)
            theNewIndex = 0;

        //Do we need to wrap?
        if (theNewIndex >= theLength)
        {
            theNewIndex = theNewIndex - theLength;

            if (itsIsRandomize)
                randomize();
        }

        //Set the index
        setIndex(theNewIndex);

        //Set the source
        setCurrentSource();

        //Write the XML source history file
        if (itsXMLHistoryFilename != null)
            writeXMLSourceHistory();
    }

    /**
     * Returns the current <code>StreamSource</code>.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the current <code>StreamSource</code>
     */
    public StreamSource currentSource()
    {
        return itsSource;
    }

    /**
     * Sets the current source to the current index entry in the source vector.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    private void setCurrentSource()
    {
        if (itsSourceVector != null)
            itsSource = (StreamSource) itsSourceVector.elementAt(itsIndex);
        else
            itsSource = null;
    }

    /**
     * "Rewinds" the current source and index by 1 and returns the new
     * source.  Wrapping calculations are done.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the new source
     */
    public StreamSource previousSource()
    {
        return previousSource(1);
    }

    /**
     * "Rewinds" the current source and index by the given amount and returns
     * the new source.  Wrapping calculations are done.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inOffset      the amount to rewind
     *
     * @return              the new source
     */
    public StreamSource previousSource(int inOffset)
    {
        setIndexAndSource(itsIndex - inOffset);
        return currentSource();
    }

    /**
     * "Fast forwards" the current source and index by 1 and returns the new
     * source.  Wrapping calculations are done.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the new source
     */
    public StreamSource nextSource()
    {
        return nextSource(1);
    }

    /**
     * "Fast forwards" the current source and index by the given amount and
     * returns the new source.  Wrapping calculations are done.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inOffset      the amount to fast forward
     *
     * @return              the new source
     */
    public StreamSource nextSource(int inOffset)
    {
        setIndexAndSource(itsIndex + inOffset);
        return currentSource();
    }

    /**
     * Sets the current source and index to the given index and
     * returns the new source.  Wrapping calculations are done.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inIndex       the index of the source to return
     *
     * @return              the source at the given index
     */
    public StreamSource indexSource(int inIndex)
    {
        setIndexAndSource(inIndex);
        return currentSource();
    }

    /**
     * Returns the source at the given index in the source vector.  This method
     * does not change the current index or source.  It just returns the source
     * from the source array.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inIndex       the index of the source to get
     *
     * @return              the source at the given index in the source vector,
     *                      <code>null</code> if the index is invalid
     */
    public StreamSource getSourceByIndex(int inIndex)
    {
        StreamSource theSource = null;

        if ((itsSourceVector != null) && (inIndex >= 0) && (inIndex < length()))
            theSource = (StreamSource) itsSourceVector.elementAt(inIndex);

        return theSource;
    }

    /**
     * Returns the index of the given <code>StreamSource</code> in the source
     * vector.  If the given source isn't found in the source vector,
     * <code>-1</code> is returned.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inSource      the <code>StreamSource</code> to get the index of
     *
     * @return              the index in the source vector, <code>-1</code> if
     *                      it isn't found
     */
    public int getIndexBySource(StreamSource inSource)
    {
        int theIndex = -1;

        if ((itsSourceVector != null) && (inSource != null))
            theIndex = itsSourceVector.indexOf(inSource);

        return theIndex;
    }

    /**
     * Adds a <code>StreamSource</code> to the end of the source vector.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inSource      the <code>StreamSource</code> to add
     */
    public void addSource(StreamSource inSource)
    {
        if ((itsSourceVector != null) && (inSource != null))
            itsSourceVector.addElement(inSource);
    }

    /**
     * Inserts a <code>StreamSource</code> at the given index of the source
     * vector.  This moves all sources after the given index up by one (that is,
     * their indexes will be one greater after the insert).  Also, if the
     * current source is after the given index, its index  will be incremented
     * by one in order to maintain consistency with the current source.  If the
     * index is invalid, nothing will be done.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inSource      the <code>StreamSource</code> to add
     * @param inIndex       the index at which to add the source
     */
    public void insertSource(StreamSource inSource, int inIndex)
    {
        if ((itsSourceVector != null) &&
            (inSource != null) &&
            (inIndex >= 0) &&
            (inIndex <= length()))
        {
            itsSourceVector.insertElementAt(inSource, inIndex);

            //Did we insert the source before the current source?
            if (itsIndex >= inIndex)
                setIndex(itsIndex + 1);
        }
    }

    /**
     * Removes a <code>StreamSource</code> from the source vector.  This moves
     * all sources after the source's index down by one (that is, their indexes
     * will be one less after the remove).  Also, if the current source is after
     * the source's index, it will be decremented by one in order to maintain
     * consistency with the current source.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inSource      the <code>StreamSource</code> to remove
     */
    public void removeSource(StreamSource inSource)
    {
        int theIndex = -1;

        if ((itsSourceVector != null) && (inSource != null))
        {
            theIndex = getIndexBySource(inSource);

            //Did we remove the element?
            if (itsSourceVector.removeElement(inSource))
            {
                //Did we remove the source before the current source?
                if (itsIndex >= theIndex)
                    setIndex(itsIndex - 1);
            }
        }
    }

    /**
     * Removes a <code>StreamSource</code> at the given index of its source
     * vector.  This moves all sources after the given index down by one (that
     * is, their indexes will be one less after the remove).  Also, if its
     * current source is after the given index, its current index will be
     * decremented by one in order to maintain consistency with its current
     * source.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inIndex      the index of the source to remove
     */
    public void removeSource(int inIndex)
    {
        if ((itsSourceVector != null) && (inIndex >= 0) && (inIndex < length()))
        {
            itsSourceVector.removeElementAt(inIndex);

            //Did we remove the source before the current source?
            if (itsIndex >= inIndex)
                setIndex(itsIndex - 1);
        }
    }

    /**
     * Re-reads its playlist file and reconstructs its source vector.  Its
     * current index is set to <code>-1</code> and randomization is done if
     * called for.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    public void rehash()
    {
        //Read the new source vector
        Vector theNewSourceVector = readPlaylist();

        if (theNewSourceVector != null)
        {
            itsSourceVector = theNewSourceVector;
            setIndex(-1);
            if (itsIsRandomize)
                randomize();
        }
    }

    /**
     * Returns the path info for the playlist file
     *
     * @author      pitboss
     * @version     1.1, 20000811
     *
     * @return      the path to the playlist file
     */
    private String getFilenamePath()
    {
        String thePath = "";
        String theFileSeparator = System.getProperty("file.separator");

        if (itsFilename != null)
        {
            int theLastSeparatorPosition =
                itsFilename.lastIndexOf(theFileSeparator);

            if (theLastSeparatorPosition != -1)
                thePath =
                    itsFilename.substring(0, theLastSeparatorPosition + 1);
        }

        return thePath;
    }

    /**
     * Reads its playlist file and returns a source vector full of
     * <code>MP3FileSource</code>s.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      a vector of <code>MP3FileSource</code>s
     */
    private Vector readPlaylist()
    {
        Vector theSourceVector = new Vector();
        String theLine = null;
        BufferedReader thePlaylistReader = null;
        String theFileSeparator = System.getProperty("file.separator");
        String thePath = getFilenamePath();

        //Read the playlist
        try
        {
            println("Reading the playlist...");

            thePlaylistReader =
                new BufferedReader
                    (new InputStreamReader(new FileInputStream(itsFilename)));

            while ((theLine = thePlaylistReader.readLine()) != null)
            {
                theLine = theLine.trim();

                //Is the line not commented out?
                if (! theLine.startsWith("#"))
                {
                    //Is this an absolute filename?
                    if ((theLine.charAt(1) == ':') ||
                        (theFileSeparator.equals(theLine.substring(0,1))))
                    {
                        println("Adding absolute file: " + theLine, 4);
                        theSourceVector.addElement
                            (new MP3FileSource(theLine));
                    }
                    else    //It's a relative path
                    {
                        println("Adding relative file: " +
                                thePath +
                                theLine,
                                4);
                        theSourceVector.addElement
                            (new MP3FileSource(thePath + theLine));
                    }
                }
            }

            println("Added " + theSourceVector.size() + " sources.");
        }
        catch (Exception e)
        {
            theSourceVector = null;
            println(e);
        }
        finally
        {
            if (thePlaylistReader != null)
            {
                try
                {
                    thePlaylistReader.close();
                }
                catch (Exception e)
                {
                    println(e);
                }
            }
        }

        return theSourceVector;
    }

    /**
     * Randmoize its source vector.  This will also set its current index to
     * <code>-1</code>.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    public void randomize()
    {
        int theLength = length();
        int theNewPosition = 0;
        Random theRandomizer = new Random();
        theRandomizer.setSeed(System.currentTimeMillis());

        if ((itsSourceVector != null) && (theLength > 1))
        {
            println("Randomizing playlist...");

            setIndex(-1);

            for (int i = 0; i < theLength; i++)
            {
                theNewPosition = i;
                while (theNewPosition == i)
                    theNewPosition =
                        (int)(theRandomizer.nextFloat() * (float)theLength);

                swapSources(i, theNewPosition);
            }

            println("Randomized playlist.");
        }
    }

    /**
     * Swap two sources in its source vector.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inIndex1      the first source index to swap
     * @param inIndex2      the second source index to swap
     */
    private void swapSources(int inIndex1, int inIndex2)
    {
        StreamSource theSource1 = getSourceByIndex(inIndex1);
        StreamSource theSource2 = getSourceByIndex(inIndex2);

        //Do we have valid indexes?
        if ((theSource1 != null) && (theSource2 != null))
        {
            //Insert the second source at the position of the first
            insertSource(theSource2, inIndex1);
            //Remove the old first source
            removeSource(inIndex1 + 1);

            //Insert the first source at the position of the second
            insertSource(theSource1, inIndex2);
            //Remove the old second source
            removeSource(inIndex2 + 1);
        }
    }

    /**
     * Writes the XML source history file.  Someone may want to rewrite this to
     * actually create an internal DOM object (so we could actually pass it on
     * to other objects if they wanted the play history) and then output it to
     * the file.  I'm just lazy.
     *
     * @author      pitboss
     * @version     1.2, 20000822
     */
    private void writeXMLSourceHistory()
    {
        PrintWriter theHistoryWriter = null;

        try
        {
            //Can we output an XML history?
            if (itsHistoryVector != null)
            {
                //Add current source to the history vector
                itsHistoryVector.insertElementAt
                    (new FilePlaylistHistoryEntry
                        ((MP3FileSource) currentSource(),
                         Logger.getDateString()),
                     0);

                //Do we need to trim the history vector?
                while (itsHistoryVector.size() > itsXMLHistorySize)
                    itsHistoryVector.removeElementAt
                        (itsHistoryVector.size() - 1);

                //Open the XML file
                theHistoryWriter =
                    new PrintWriter(new FileWriter(itsXMLHistoryFilename),
                                    true);

                //Write the header
                theHistoryWriter.println("<?xml version=\"1.0\"?>");
                theHistoryWriter.println("<!DOCTYPE " +
                                         FILE_PLAYLIST_HISTORY_ELEMENT +
                                         " SYSTEM \"" +
                                         FILE_PLAYLIST_HISTORY_DTD +
                                         "\">");

                //Write the root element
                theHistoryWriter.println("<" +
                                         FILE_PLAYLIST_HISTORY_ELEMENT +
                                         ">");

                //Loop through the history
                int theSize = itsHistoryVector.size();
                for (int i = 0; i < theSize; i++)
                {
                    writeXMLSource(theHistoryWriter,
                                   (FilePlaylistHistoryEntry)
                                        itsHistoryVector.elementAt(i));
                }

                //Close the root element
                theHistoryWriter.println("</" +
                                         FILE_PLAYLIST_HISTORY_ELEMENT +
                                         ">");
            }
        }
        catch (Exception e)
        {
            println(e);
        }
        finally
        {
            if (theHistoryWriter != null)
                theHistoryWriter.close();
        }
    }

    /**
     * Writes the XML info for a source.  Someone may want to rewrite this to
     * actually create an internal DOM object (so we could actually pass it on
     * to other objects if they wanted the play history) and then output it to
     * the file.  I'm just lazy.
     *
     * @author      pitboss
     * @version     1.2, 20000822
     *
     * @param inXMLFile     the <code>PrintWriter</code> for the history file
     * @param inEntry       the <code>FilePlaylistHistoryEntry</code> containing
     *                      the source to write
     */
    private void writeXMLSource(PrintWriter inXMLFile,
                                FilePlaylistHistoryEntry inEntry)
    {
        String theString = null;
        MP3FileSource theSource = inEntry.itsSource;

        //Is it a valid source?
        if (theSource != null)
        {
            //Write the head tag and time attribute
            inXMLFile.println("\t<" +
                              MP3_FILE_SOURCE_ELEMENT +
                              " " +
                              TIME_ATTRIBUTE +
                              "=\"" +
                              inEntry.itsDateString +
                              "\">");

            //Get the filename
            theString = theSource.name();
            if (theString != null)
            {
                inXMLFile.println("\t\t<" + FILENAME_ELEMENT + ">");
                inXMLFile.println("\t\t\t" + theString);
                inXMLFile.println("\t\t</" + FILENAME_ELEMENT + ">");
            }

            //Get the title
            theString = theSource.title();
            if (theString != null)
            {
                inXMLFile.println("\t\t<" + TITLE_ELEMENT + ">");
                inXMLFile.println("\t\t\t" + theString);
                inXMLFile.println("\t\t</" + TITLE_ELEMENT + ">");
            }

            //Get the artist
            theString = theSource.artist();
            if (theString != null)
            {
                inXMLFile.println("\t\t<" + ARTIST_ELEMENT + ">");
                inXMLFile.println("\t\t\t" + theString);
                inXMLFile.println("\t\t</" + ARTIST_ELEMENT + ">");
            }

            //Get the album
            theString = theSource.album();
            if (theString != null)
            {
                inXMLFile.println("\t\t<" + ALBUM_ELEMENT + ">");
                inXMLFile.println("\t\t\t" + theString);
                inXMLFile.println("\t\t</" + ALBUM_ELEMENT + ">");
            }

            //Write the length in seconds
            inXMLFile.println("\t\t<" + LENGTH_IN_SECONDS_ELEMENT + ">");
            inXMLFile.println("\t\t\t" + theSource.lengthInSeconds());
            inXMLFile.println("\t\t</" + LENGTH_IN_SECONDS_ELEMENT + ">");

            //Write the close head tag
            inXMLFile.println("\t</" + MP3_FILE_SOURCE_ELEMENT + ">");
        }
    }
}
