/*
JRev: A JAVA MP3 Source Streamer
Copyright (C) 2000  Thomas Bohmbach, Jr.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package org.jrev;

import java.io.BufferedReader;
import java.io.DataOutputStream;
import java.io.InputStreamReader;
import java.net.Socket;

/**
 * The streaming "engine" class.  At its most basic level, it calls
 * <code>currentSource()</code> from its {@link Playlist}, streams that
 * source to its destination, and when complete, calls <code>nextSource()</code>
 * and repeats the process until it is told to stop or it looses the connection
 * to its destination server.
 *
 * @author      pitboss
 * @version     1.0, 20000728
 * @since       1.0
 * @see         Playlist
 * @see         Logger
 */
public class JRev extends Thread
{
    private static final String VERSION = "1.0.1";
    private static final String CONFIG_FILE = "jrev_config.xml";
    private static final int MAX_NUM_RETRIES = 5;
    private static final int MAX_CONNECT_RETRIES = 10;
    private static final String SERVERRESPONSE_MOUNT_POINT_TAKEN
        = "ERROR - Mount Point Taken or Invalid";

    private static final int CONNECT_OK   = 0;
    private static final int CONNECT_FAIL = 1;
    private static final int CONNECT_BUSY = 2;

    private String itsName = null;
    private String itsDescription = null;
    private String itsURL = null;
    private String itsGenre = null;
    private int itsBitrate = 128;
    private String itsServer = null;
    private int itsServerPort = 8000;
    private String itsServerPassword = null;
    private String itsMountPoint = null;
    private boolean itsIsPublic = false;
    private boolean itsUseIcyHeaders = true;
    private int itsBufferSizeInBytes = 4000;
    private byte itsBuffer[] = null;
    private int itsDelayConstant = (itsBufferSizeInBytes * 8) / itsBitrate;
    private int itsOverhead = 5;
    private LoggerClient itsLoggerClient = null;
    private Playlist itsPlaylist = null;
    private Thread itsThread = null;
    private Socket itsServerSocket = null;
    private BufferedReader itsServerReader = null;
    private DataOutputStream itsServerOutputStream = null;
    private StreamSource itsSource = null;
    private boolean itsIsRunning = true;
    private boolean itsSkipFlag = false;
    private boolean itsStopFlag = false;
    private boolean itsRestartFlag = true;  //Note: JRev will not run if this
                                            //is set to false initially

    /**
     * Creates an instance of <code>JRev</code> with a mountpoint.
     *
     * @author      Bryan Buchanan
     * @version     1.2, 20000822
     *
     * @param inName            the name of this stream
     * @param inDescripion      the description of this stream
     * @param inURL             the URL of this stream
     * @param inGenre           the genre of this stream
     * @param inBitrate         the bitrate that this stream will be using
     * @param inServer          the server to stream to.  This can be a DNS
     *                          name or an IP address.
     * @param inPort            the port on the destination server to stream to
     * @param inPassword        the password for the destination server
     * @param inMountPoint      the mount point for the destination server
     * @param inIsPublic        determines if this a public stream and if it
     *                          should it register with public icecast location
     *                          servers
     * @param inUseIcyHeaders   determines whether to send ICY headers to the
     *                          server or x-audiocast headers
     * @param inBuffersize      the internal buffersize in bytes to use. May be
     *                          <code>null</code> for default size.
     * @param inOverhead        the overhead factor to use.  May be
     *                          <code>null</code> for default overhead.
     * @param inLogger          the log settings to use.  May be
     *                          <code>null</code> for no logging
     * @param inPlaylist        the {@link Playlist} that this class will
     *                          use to get its {@link StreamSource}
     *
     * @see     Logger
     * @see     Playlist
     * @see     StreamSource
     */
    public JRev(String inName,
                String inDescription,
                String inURL,
                String inGenre,
                int inBitrate,
                String inServer,
                int inPort,
                String inPassword,
                String inMountPoint,
                boolean inIsPublic,
                boolean inUseIcyHeaders,
                Integer inBufferSize,
                Integer inOverhead,
                Logger inLogger,
                Playlist inPlaylist)
    {
        itsName = inName;
        itsDescription = inDescription;
        itsURL = inURL;
        itsGenre = inGenre;
        itsServer = inServer;
        itsServerPort = inPort;
        itsServerPassword = inPassword;
        itsMountPoint = inMountPoint;
        itsIsPublic = inIsPublic;
        itsUseIcyHeaders = inUseIcyHeaders;

        //Set the bitrate (and delay constant)
        setBitrate(inBitrate);

        //Optional parameter (Default: 4000)
        if (inBufferSize != null)
            setBuffersize(inBufferSize.intValue());  //And delay constant
        else
            setBuffersize(itsBufferSizeInBytes);

        //Optional parameter (Default: 5)
        if (inOverhead != null)
            itsOverhead = inOverhead.intValue();

        itsLoggerClient = new LoggerClient(inLogger);

        itsPlaylist = inPlaylist;

        itsThread = new Thread(this);
        itsThread.start();
    }

    /**
     * Creates an instance of <code>JRev</code> with no mountpoint.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inName            the name of this stream
     * @param inDescripion      the description of this stream
     * @param inURL             the URL of this stream
     * @param inGenre           the genre of this stream
     * @param inBitrate         the bitrate that this stream will be using
     * @param inServer          the server to stream to.  This can be a DNS
     *                          name or an IP address.
     * @param inPort            the port on the destination server to stream to
     * @param inPassword        the password for the destination server
     * @param inIsPublic        determines if this a public stream and if it
     *                          should it register with public icecast location
     *                          servers
     * @param inUseIcyHeaders   determines whether to send ICY headers to the
     *                          server or x-audiocast headers
     * @param inBuffersize      the internal buffersize in bytes to use. May be
     *                          <code>null</code> for default size.
     * @param inOverhead        the overhead factor to use.  May be
     *                          <code>null</code> for default overhead.
     * @param inLogger          the log settings to use.  May be
     *                          <code>null</code> for no logging
     * @param inPlaylist        the {@link Playlist} that this class will
     *                          use to get its {@link StreamSource}
     *
     * @see     Logger
     * @see     Playlist
     * @see     StreamSource
     */
    public JRev(String inName,
                String inDescription,
                String inURL,
                String inGenre,
                int inBitrate,
                String inServer,
                int inPort,
                String inPassword,
                boolean inIsPublic,
                boolean inUseIcyHeaders,
                Integer inBufferSize,
                Integer inOverhead,
                Logger inLogger,
                Playlist inPlaylist)
    {
        itsName = inName;
        itsDescription = inDescription;
        itsURL = inURL;
        itsGenre = inGenre;
        itsServer = inServer;
        itsServerPort = inPort;
        itsServerPassword = inPassword;
        itsIsPublic = inIsPublic;
        itsUseIcyHeaders = inUseIcyHeaders;

        //Set the bitrate (and delay constant)
        setBitrate(inBitrate);

        //Optional parameter (Default: 4000)
        if (inBufferSize != null)
            setBuffersize(inBufferSize.intValue());  //And delay constant
        else
            setBuffersize(itsBufferSizeInBytes);

        //Optional parameter (Default: 5)
        if (inOverhead != null)
            itsOverhead = inOverhead.intValue();

        itsLoggerClient = new LoggerClient(inLogger);

        itsPlaylist = inPlaylist;

        itsThread = new Thread(this);
        itsThread.start();
    }

    /**
     * Returns whether this <code>JRev</code> is running.
     *
     * @author      pitboss
     * @version     1.1, 20000813
     *
     * @return      <code>true</code> if this <code>JRev</code> instance is
     *              runnning, <code>false</code> otherwise.
     */
    public boolean isRunning()
    {
        return itsIsRunning;
    }

    /**
     * Stops this <code>JRev</code> instance and shuts down the streaming
     * thread.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    public void shutdown()
    {
        itsLoggerClient.println("Shutting down " + itsName + "...");

        itsStopFlag = true;
        itsRestartFlag = false;
    }

    /**
     * Stops this <code>JRev</code> instance, closes its connection to the
     * server, and then reconnects.
     *
     * @author      pitboss
     * @version     1.1, 20000813
     */
    public void restart()
    {
        itsLoggerClient.println("Restarting " + itsName + "...");

        itsStopFlag = true;
        itsRestartFlag = true;
    }

    /**
     * Skips the current <code>StreamSource</code> and moves on to the next.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    public void skip()
    {
        itsSkipFlag = true;
    }

    /**
     * Gets the name of this stream.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the name of the stream
     */
    public String name()
    {
        return itsName;
    }

    /**
     * Gets the current source that is being streamed.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the currently streaming <code>StreamSource</code>
     */
    public StreamSource getSource()
    {
        return itsSource;
    }

    /**
     * Gets the bitrate that this stream is being delivered at.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the stream bitrate
     */
    public int getBitrate()
    {
        return itsBitrate;
    }

    /**
     * Sets the bitrate that this stream is being delivered at.  Trying to
     * change this mid-stream may cause problems.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inBitrate     the bitrate to use.
     */
    private void setBitrate(int inBitrate)
    {
        if (itsBitrate != 0)
        {
            itsBitrate = inBitrate;

            itsDelayConstant = (itsBufferSizeInBytes * 8) / itsBitrate;
        }
    }

    /**
     * Gets the <code>Playlist</code> that this stream is using to get its
     * <code>StreamSource</code>s.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the current <code>Playlist</code>
     */
    public Playlist getPlaylist()
    {
        return itsPlaylist;
    }

    /**
     * Sets the <code>Playlist</code> that this stream is using to get its
     * <code>StreamSource</code>s.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inPlaylist    the new <code>Playlist</code>
     */
    public void setPlaylist(Playlist inPlaylist)
    {
        itsPlaylist = inPlaylist;
    }

    /**
     * Sets the buffer size (in bytes) that will be used for streaming data to
     * the destination server.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inBuffersize  the buffer size in bytes
     */
    private void setBuffersize(int inBuffersize)
    {
        itsBufferSizeInBytes = inBuffersize;
        itsBuffer = new byte[itsBufferSizeInBytes];
        itsDelayConstant = (itsBufferSizeInBytes * 8) / itsBitrate;
    }

    /**
     * Gets the <code>Logger</code> that will be used for logging.
     * <p>
     * Note that this method only exists because we can't extend
     * {@link LoggerClient} (because we already extend <code>Thread</code>) and
     * therefore must implement <code>LoggerClient</code> finctionality by
     * delegation.  This is why Java really misses multiple inheritance.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the current <code>Logger</code> that this
     *              <code>JRev</code> is using to log output
     */
    public Logger getLogger()
    {
        Logger theLogger = null;

        if (itsLoggerClient != null)
            theLogger = itsLoggerClient.getLogger();

        return theLogger;
    }

    /**
     * Gets the current <code>StreamSource</code> from the current
     * <code>Playlist</code>.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the current <code>StreamSource</code>.  Can be
     *              <code>null</code> if its <code>Playlist</code> is
     *              <code>null</code> or its <code>Playlist</code> has no
     *              <code>StreamSource</code>s.
     */
    private StreamSource getCurrentSourceFromPlaylist()
    {
        StreamSource theCurrentSource = null;

        //Do we have a playlist?
        if (itsPlaylist != null)
            theCurrentSource = itsPlaylist.currentSource();

        return theCurrentSource;
    }

    /**
     * Gets the next <code>StreamSource</code> from the current
     * <code>Playlist</code>.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the next <code>StreamSource</code>.  Can be
     *              <code>null</code> if its <code>Playlist</code> is
     *              <code>null</code> or its <code>Playlist</code> has no
     *              <code>StreamSource</code>s left.
     */
    private StreamSource getNextSourceFromPlaylist()
    {
        StreamSource theNextSource = null;

        //Do we have a playlist?
        if (itsPlaylist != null)
            theNextSource = itsPlaylist.nextSource();

        return theNextSource;
    }

    /**
     * Sends a <code>String</code> to the destination server.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inString      the <code>String</code> to send to the server
     *
     * @return              <code>true</code> if the <code>String</code> was
     *                      sent successfully, <code>false</code> otherwise
     */
    private boolean stream(String inString)
    {
        boolean theIsSuccess = false;

        try
        {
            itsServerOutputStream.writeBytes(inString);
            itsServerOutputStream.flush();
            theIsSuccess = true;
        }
        catch (Exception e)
        {
            itsLoggerClient.println(e);
        }

        return theIsSuccess;
    }

    /**
     * Sends an array of bytes to the destination server.  This is the method
     * that actually sends the data buffer to the server.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inByteArray[]     the array of bytes to send to the server
     * @param inLength          the number of bytes to send
     *
     * @return                  <code>true</code> if the array of bytes were
     *                          sent successfully, <code>false</code> otherwise
     */
    private boolean stream(byte inByteArray[], int inLength)
    {
        boolean theIsSuccess = false;

        try
        {
            itsServerOutputStream.write(inByteArray, 0, inLength);
            itsServerOutputStream.flush();
            theIsSuccess = true;
        }
        catch (Exception e)
        {
            itsLoggerClient.println(e);
        }

        return theIsSuccess;
    }

    /**
     * Sends the current <code>StreamSource</code> to the destination server.
     * This method contains Dormondo's "magic loop."  It will attempt to send
     * the whole source (until the source is out of bytes) unless it is told to
     * skip or stop altogether.  If it is interrupted, it will finish sending
     * the current buffer of data, and then stop.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    private void stream()
    {
        long theExtra = 0;
        long theTime = 0;
        long theOldTime = 0;
        long theDelay = 0;
        int theNumBytesRead = 0;
        int theNumRetries = 0;

        try
        {
            //Can we open our source?
            if ((itsSource != null) &&
                itsSource.open() &&
                (itsBuffer != null) &&
                !itsStopFlag)
            {
                String theTitle = itsSource.title();
                if (theTitle == null)
                    theTitle = itsSource.name();
                if (theTitle == null)
                    theTitle = "an unknown source";

                itsLoggerClient.println("Streaming " + theTitle + "...");

                //Get the initial time
                theOldTime = System.currentTimeMillis();

                //Read buffers of data until there's no more to read
                while (((theNumBytesRead = itsSource.read(itsBuffer)) != -1) &&
                       !itsSkipFlag &&
                       !itsStopFlag)
                {
                    //Send buffer to server
                    while ((! stream(itsBuffer, theNumBytesRead)) &&
                           (theNumRetries < MAX_NUM_RETRIES))
                        theNumRetries++;

                    //Was there a problem sending the data?
                    if (theNumRetries < MAX_NUM_RETRIES)    //Small problem
                        theNumRetries = 0;  //Reset for the next time
                    else    //There is no next time, this is fatal
                        restart();  //But lets try again

                    //Get the time now (after reading and sending the buffer)
                    theTime = System.currentTimeMillis();

                    //Calculate the new delay
                    theDelay = ( ( itsDelayConstant -
                                   ( theTime - theOldTime ) ) -
                                 itsOverhead ) -
                               theExtra;

                    //Reset the amount behind
                    theExtra = 0;

                    //Are we behind?
                    if (theDelay < 0)
                    {
                        theExtra = Math.abs(theDelay);
                        theDelay = 0;
                    }

                    //Delay
                    Thread.sleep(theDelay);

                    //Reset theOldTime for next buffer of data
                    theOldTime = System.currentTimeMillis();
                }

                itsLoggerClient.println("Streamed " + itsSource.title() + ".", 2);
            }
        }
        catch (Exception e)
        {
            itsLoggerClient.println(e);
        }
        finally
        {
            //Reset the skip flag
            itsSkipFlag = false;

            if (itsSource != null)
            {
                itsLoggerClient.println("Closing " + itsSource.title() + "...", 2);
                itsSource.close();
                itsLoggerClient.println("Closed " + itsSource.title() + ".", 2);
            }
        }
    }

    /**
     * Connects and logs in this <code>JRev</code> instance to the
     * destination server.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      an int constant indicating the status.  CONNECT_OK
     * indicates a successful connection.  CONNECT_FAIL indicates a failed
     * connection.  CONNECT_BUSY indicates a connection that failed because
     * the mountpoint was still in use.  A retry may be possible in this
     * situation depending on server.
     */
    private int connectToServer()
    {
        int theIsSuccess = CONNECT_FAIL;

    	//Connect to the server
        while (itsServerSocket == null)
        {
            try
            {
                //Create the connect string
                String theString = new String("Connecting to " +
                                              itsServer +
                                              " on port " +
                                              itsServerPort);

                //Do we have a mountpoint?
                if ((itsMountPoint != null) &&
                    (! itsMountPoint.equalsIgnoreCase("default")))
                    theString += " at mountpoint " + itsMountPoint;

                theString += "...";

                itsLoggerClient.println(theString);

                itsServerSocket = new Socket(itsServer, itsServerPort);
                itsServerReader =
                    new BufferedReader
                     (new InputStreamReader(itsServerSocket.getInputStream()));
                itsServerOutputStream =
                    new DataOutputStream(itsServerSocket.getOutputStream());
            }
            catch (Exception e)
            {
            	itsServerSocket = null;

                itsLoggerClient.println(e);

                try
                {
                    //Sleep for 5 seconds before trying again
                    Thread.sleep(5000);
              	}
             	catch (InterruptedException ie)
                {
                }
            }
     	}

        //Send the password
        //Did we set a mountpoint?
        if (itsMountPoint != null)
        {
            itsLoggerClient.println("Sending password and mountpoint...", 2);

            stream("SOURCE " +
                   itsServerPassword +
                   " /" +
                   itsMountPoint +
                   "\n\n");

            itsLoggerClient.println("Sent password and mountpoint.", 2);
        }
        else    //We didn't set a mountpoint
        {
            itsLoggerClient.println("Sending password...", 2);
            stream(itsServerPassword + "\n");
            itsLoggerClient.println("Sent password.", 2);
        }

        try
        {
            String theResponseString = null;
            int theNumTries = 0;
            
            //Read the response
            while ((theNumTries < 10) && (theResponseString == null))
            {
                theResponseString = itsServerReader.readLine();
                
                theNumTries++;
                
                if (theNumTries >= 10)
                    theResponseString = "No response";
            }
            
            theResponseString.trim();
            
            itsLoggerClient.println("The server responded with: " +
                                    theResponseString, 2);
            
            //Are we connected?
            if (theResponseString.startsWith("OK"))
            {
                itsLoggerClient.println("Connected to " + itsServer + ".");
                theIsSuccess = CONNECT_OK;
            }
            else if (theResponseString.startsWith(SERVERRESPONSE_MOUNT_POINT_TAKEN))
            {
                itsLoggerClient.println("Mount point busy.");
                theIsSuccess = CONNECT_BUSY;
            }
            else
            {
                itsLoggerClient.println("Bad Login: The server responded with: " +
                                        theResponseString);
                shutdown();
                theIsSuccess = CONNECT_FAIL;
            }
        }
        catch (Exception e)
        {
            itsLoggerClient.println(e);
            theIsSuccess = CONNECT_FAIL;
        }
        finally
        {
            //Free up everything if we didn't succeed.
            if ( theIsSuccess != CONNECT_OK ) {
                disconnectFromServer();
            }
        }
        return theIsSuccess;
    }

    /**
     * Disconnects this <code>JRev</code> instance from the destination
     * server.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    private void disconnectFromServer()
    {
        itsLoggerClient.println("Disconnecting from " + itsServer + "...");

        try
        {
            if (itsServerOutputStream != null)
            {
                itsLoggerClient.println("Closing output stream...", 2);
                itsServerOutputStream.close();
                itsLoggerClient.println("Closed output stream.", 2);
            }
        }
        catch (Exception e)
        {
            itsLoggerClient.println(e);
        }
        finally
        {
            itsServerOutputStream = null;
        }

        try
        {
            if (itsServerReader != null)
            {
                itsLoggerClient.println("Closing input reader...", 2);
                itsServerReader.close();
                itsLoggerClient.println("Closed input reader.", 2);
            }
        }
        catch (Exception e)
        {
            itsLoggerClient.println(e);
        }
        finally
        {
            itsServerReader = null;
        }

        try
        {
            if (itsServerSocket != null)
            {
                itsLoggerClient.println("Closing socket...", 2);
                itsServerSocket.close();
                itsLoggerClient.println("Closed socket.", 2);
            }
        }
        catch (Exception e)
        {
            itsLoggerClient.println(e);
        }
        finally
        {
            itsServerSocket = null;
        }
    }

    /**
     * Sends ICY headers to the destination server.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    private void sendIcyHeaders()
    {
        itsLoggerClient.println("Sending icy headers...", 2);

        stream("icy-name: " + itsName + "\n");
        stream("icy-genre: " + itsGenre + "\n");
        stream("icy-url: " + itsURL + "\n");
        if (itsIsPublic)
            stream("icy-pub: yes\n");
        else
            stream("icy-pub: no\n");
        stream("icy-br: " + itsBitrate + "\n");
        stream("icy-desc: " + itsDescription + "\n");
        stream("\n");

        itsLoggerClient.println("Sent icy headers.", 2);
    }

    /**
     * Sends x-audiocast headers to the destination server.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    private void sendXAudiocastHeaders()
    {
        itsLoggerClient.println("Sending XAudiocast headers...", 2);

        stream("x-audiocast-name: " + itsName + "\n");
        stream("x-audiocast-description: " + itsDescription + "\n");
        stream("x-audiocast-url: " + itsURL + "\n");
        stream("x-audiocast-genre: " + itsGenre + "\n");
        stream("x-audiocast-bitrate: " + itsBitrate + "\n");
        if (itsIsPublic)
            stream("x-audiocast-public: yes\n");
        else
            stream("x-audiocast-public: no\n");
        stream("\n");

        itsLoggerClient.println("Sent XAudiocast headers.", 2);
    }

    /**
     * The main <code>Thread</code> method.  Connects to the destination server,
     * sends the headers, and starts streaming sources.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    public void run()
    {
        itsIsRunning = true;

        while (itsRestartFlag)
        {
            //Reset the flags
            itsRestartFlag = false;
            itsStopFlag = false;
            int theConnectStatus = CONNECT_FAIL;
            int theRetryCount = 0;
            long theBusyWaitMillis = 100;

            while (!itsStopFlag)
            {
                try
                {
                    //Connect to the server
                    while ( true ) {
                        theConnectStatus = connectToServer();
                        if ( theConnectStatus == CONNECT_OK ) break;
                        if ( theConnectStatus == CONNECT_FAIL ) return;
                        if ( theConnectStatus == CONNECT_BUSY ) {
                            if ( (++theRetryCount) > MAX_CONNECT_RETRIES ) {
                                itsLoggerClient.println("Max retries exceeded, bailing out.");
                                return;
                            }
                            itsLoggerClient.println("Sleeping for " + theBusyWaitMillis 
                                                    + " ms, then retrying.");
                            try { Thread.sleep(theBusyWaitMillis); } catch ( Exception e ) {}
                            theBusyWaitMillis *= 3;
                            theBusyWaitMillis /= 2;
                        }
                    }

                   	if (theConnectStatus == CONNECT_OK)
                    {
                        //Send the headers
                        if (itsUseIcyHeaders)
                            sendIcyHeaders();
                        else
                            sendXAudiocastHeaders();

                        //Get the first source
                        itsSource = getCurrentSourceFromPlaylist();

                        //Loop through the playlist and stream each source
                        while (!itsStopFlag)
                        {
                            //Stream the source to the server
                            stream();

                            //Do we need another source?
                            if (!itsStopFlag)
                                itsSource = getNextSourceFromPlaylist();
                            else
                                itsSource = null;
                        }
                    } else {
                        
                    }
                }
                catch (Exception e)
                {
                    itsLoggerClient.println(e);
                }
                finally
                {
                    if (itsSource != null)
                        itsSource.close();

                    if (itsServerSocket != null)
                        disconnectFromServer();
                }
            }
        }

        itsIsRunning = false;
    }

    /**
     * Provides "out-of-the-box" functionality.  It will read from a config file
     * to set up a <code>Playlist</code>, a <code>JRev</code> instance, and
     * optionally a <code>TelnetRemoteInterface</code> to control them both.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param args  the array of command-line args
     */
    public static void main(String[] args)
    {
        try
        {
            String theConfigFilename = CONFIG_FILE;

            // is there anything to do?
            if (args.length != 0)
            {
                // check parameters
                for (int i = 0; i < args.length; i++)
                {
                    String theArg = args[i];

                    //Options
                    if (theArg.startsWith("-"))
                    {
                        //Config filename
                        if (theArg.equals("-c"))
                        {
                            if (i == args.length - 1)
                            {
                                System.err.println("Error: missing config filename");
                                System.exit(1);
                            }
                            theConfigFilename = args[++i];
                            continue;
                        }

                        //Print the usage text
                        if (theArg.equals("-h"))
                        {
                            printUsage();
                            System.exit(1);
                        }
                    }
                }
            }

            //Create the config file and needed classes
            ConfigFile theConfigFile = new ConfigFile(theConfigFilename);
            FilePlaylist thePlaylist = null;
            JRev theStreamer = null;
            TelnetRemoteInterface theTRI = null;

            //Was the config file able to complete its reading?
            if (theConfigFile.complete())
            {
                //Create the playlist
                Logger thePlaylistLogger = new Logger(theConfigFile.getIsPlaylistQuiet(),
                                                      theConfigFile.getPlaylistLogFilename(),
                                                      theConfigFile.getPlaylistOutputLevel());

                //Are we keeping a history file?
                if (theConfigFile.getHasFilePlaylistHistory())
                {
                    thePlaylist = new FilePlaylist(theConfigFile.getPlaylistFilename(),
                                                   theConfigFile.getIsRandomizePlaylist(),
                                                   thePlaylistLogger,
                                                   theConfigFile.getPlaylistHistoryFilename(),
                                                   theConfigFile.getPlaylistHistorySize());
                }
                else    //We don't want to keep a history
                {
                    thePlaylist = new FilePlaylist(theConfigFile.getPlaylistFilename(),
                                                   theConfigFile.getIsRandomizePlaylist(),
                                                   thePlaylistLogger);
                }

                //Create the streamer
                Logger theJRevLogger = new Logger(theConfigFile.getIsJRevQuiet(),
                                                  theConfigFile.getJRevLogFilename(),
                                                  theConfigFile.getJRevOutputLevel());
                theStreamer = new JRev(theConfigFile.getName(),
                                       theConfigFile.getDescription(),
                                       theConfigFile.getURL(),
                                       theConfigFile.getGenre(),
                                       theConfigFile.getBitrate(),
                                       theConfigFile.getServer(),
                                       theConfigFile.getJRevPort(),
                                       theConfigFile.getJRevPassword(),
                                       theConfigFile.getJRevMountPoint(),
                                       theConfigFile.getIsPublic(),
                                       theConfigFile.getIsIcyHeaders(),
                                       new Integer(theConfigFile.getBuffersize()),
                                       new Integer(theConfigFile.getOverhead()),
                                       theJRevLogger,
                                       thePlaylist);

                //Create the Telnet Remote Interface (TRI)
                if (theConfigFile.getHasTRI())
                {
                    Logger theTRILogger = new Logger(theConfigFile.getIsTRIQuiet(),
                                                     theConfigFile.getTRILogFilename(),
                                                     theConfigFile.getTRIOutputLevel());
                    theTRI = new TelnetRemoteInterface(theConfigFile.getTRIPort(),
                                                       theConfigFile.getTRIPassword(),
                                                       theStreamer,
                                                       thePlaylist,
                                                       theTRILogger);
                }
            }
            else
            {
                System.out.println("Error reading config file (" + theConfigFilename + ").");
            }
        }
        catch (Exception e)
        {
            System.out.println(e);
        }
    }

    /**
     * Prints the "out-of-the-box" functionality usage signature.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    private static void printUsage()
    {
        System.out.println(" ");
        System.out.println("JRev - Version " + VERSION);
        System.out.println(" ");
        System.out.println("Usage:");
        System.out.println("java org.jrev.JRev [-c <configfile>] [-h]");
        System.out.println(" ");
        System.out.println("     Options explained (compiled default in parenthesis):");
        System.out.println("     -c: Configuration file to use (" + CONFIG_FILE + ")");
        System.out.println("     -h: Print this usage text");
        System.out.println(" ");
        System.out.println(" ");
    }
}
