/*
JRev: A JAVA MP3 Source Streamer
Copyright (C) 2000  Thomas Bohmbach, Jr.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package org.jrev;

import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.Date;
import java.text.SimpleDateFormat;

/**
 * A class to handle logging to stdout and/or to a log file.  Additionally, a
 * debug level can be set to limit the amount of output logged.
 *
 * @author      pitboss
 * @version     1.0, 20000728
 * @since       1.0
 */
public class Logger
{
    static SimpleDateFormat itsDateFormat =
        new SimpleDateFormat("yyyyMMdd HH:mm:ss z");

    boolean itsIsQuiet = true;
    PrintWriter itsLogPrintWriter = null;

    int itsDebugLevel = 0;

    /**
     * Creates a simple instance of a <code>Logger</code> that doesn't log
     * anywhere.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    public Logger()
    {
    }

    /**
     * Creates an instance of a <code>Logger</code> that may output to stdout,
     * but doesn't have a log file associated with it.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inIsQuiet     if <code>true</code> the logger will be quiet (that
     *                      is, it won't log to stdout); if <code>false</code>
     *                      the logger will output to stdout
     */
    public Logger(boolean inIsQuiet)
    {
        itsIsQuiet = inIsQuiet;
    }

    /**
     * Creates an instance of a <code>Logger</code> that may output to stdout
     * and a log file.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inIsQuiet         if <code>true</code> the logger will be quiet
     *                          (that is, it won't log to stdout); if
     *                          <code>false</code> the <code>Logger</code> will
     *                          output to stdout
     * @param inLogFilename     the file name to log output to; if
     *                          <code>null</code>, the <code>Logger</code> won't
     *                          output to a log file.
     */
    public Logger(boolean inIsQuiet, String inLogFilename)
    {
        itsIsQuiet = inIsQuiet;

        if (inLogFilename != null)
            initPrintWriter(inLogFilename);
    }

    /**
     * Creates an instance of a <code>Logger</code> that may output to stdout
     * and a log file with a given debug level.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inIsQuiet         if <code>true</code> the logger will be quiet
     *                          (that is, it won't log to stdout); if
     *                          <code>false</code> the <code>Logger</code> will
     *                          output to stdout
     * @param inLogFilename     the file name to log output to; if
     *                          <code>null</code>, the <code>Logger</code> won't
     *                          output to a log file.
     * @param inDebugLevel      the debug level of this logger; <code>0</code>
     *                          is normal; larger numbers indicate more verbose
     *                          output
     */
    public Logger(boolean inIsQuiet, String inLogFilename, int inDebugLevel)
    {
        itsIsQuiet = inIsQuiet;
        itsDebugLevel = inDebugLevel;

        if (inLogFilename != null)
            initPrintWriter(inLogFilename);
    }

    /**
     * Creates its PrintWriter with AutoFlush and Append for its log file.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inLogFilename     the file name to use for the log file
     */
    private void initPrintWriter(String inLogFilename)
    {
        try
        {
            itsLogPrintWriter =
                new PrintWriter(new FileWriter(inLogFilename, true), true);
            itsLogPrintWriter.println();
            itsLogPrintWriter.println();
        }
        catch (Exception e)
        {
            println(e);
        }
    }

    /**
     * Cleans up this <code>Logger</code> upon garbage collection.  It will close
     * the log file, if any.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    protected void finalize()
    {
        try
        {
            if (itsLogPrintWriter != null)
                itsLogPrintWriter.close();
        }
        catch (Exception e)
        {
        }

        itsLogPrintWriter = null;
    }

    /**
     * Gets the debug level for this <code>Logger</code>.  By default, ouput is
     * assumed to be at a debug level of <code>0</code> unless otherwise
     * specified.  That means, as long as the debug level for this
     * <code>Logger</code> is greater than or equal to <code>0</code> the output
     * will be logged.  Thus, the higher the debug level, the more verbose the
     * logging will be.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the debug level
     */
    public int getDebugLevel()
    {
        return itsDebugLevel;
    }

    /**
     * Sets the debug level for this <code>Logger</code>.  By default, ouput is
     * assumed to be at a debug level of <code>0</code> unless otherwise
     * specified.  That means, as long as the debug level for this
     * <code>Logger</code> is greater than or equal to <code>0</code> the output
     * will be logged.  Thus, the higher the debug level, the more verbose the
     * logging will be.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inDebugLevel      the new debug level
     */
    public void setDebugLevel(int inDebugLevel)
    {
        itsDebugLevel = inDebugLevel;
    }

    /**
     * Gets the date string that is prefixed to any <code>println</code> output.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the date string
     */
    static public String getDateString()
    {
        String theString = "";
        Date theCurrentDate = new Date();

        if (itsDateFormat != null)
            theString = new String(itsDateFormat.format(theCurrentDate));

        return theString;
    }

    /**
     * Outputs the given string to the logs.  No prefix date or newline will be
     * output.  Use {@link #println(String)} if you need a prefix date or
     * automatic newline.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inString  the <code>String</code> to output
     */
    public void print(String inString)
    {
        try
        {
            //Do we output to System.out?
            if (!itsIsQuiet)
            {
                System.out.print(inString);
                System.out.flush();
            }

            //Do we output to a log file?
            if (itsLogPrintWriter != null)
            {
                itsLogPrintWriter.print(inString);
                itsLogPrintWriter.flush();
            }
        }
        catch (Exception e)
        {
        }
    }

    /**
     * Outputs the given string to the logs only if the passed in debug level is
     * less than or equal to its debug level.  No prefix date or newline will be
     * output.  Use {@link #println(String, int)} if you need a prefix date or
     * automatic newline.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inString      the <code>String</code> to output
     * @param inDebugLevel  the debug level of this output (normal is
     *                      <code>0</code>)
     */
    public void print(String inString, int inDebugLevel)
    {
        if (inDebugLevel <= itsDebugLevel)
            print(inString);
    }

    /**
     * Outputs a newline to the logs.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    public void println()
    {
        try
        {
            //Do we output to System.out?
            if (!itsIsQuiet)
                System.out.println();

            //Do we output to a log file?
            if (itsLogPrintWriter != null)
                itsLogPrintWriter.println();
        }
        catch (Exception e)
        {
        }
    }

    /**
     * Outputs the given string to the logs, including a prefix date and
     * trailing newline.  Use {@link #print(String)} if you don't want a prefix
     * date and automatic newline.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inString  the <code>String</code> to output
     */
    public void println(String inString)
    {
        try
        {
            //Prepend the date
            Date theCurrentDate = new Date();
            String theStringToPrint = new String("[" +
                                                 getDateString() +
                                                 "] " +
                                                 inString);

            //Do we output to System.out?
            if (!itsIsQuiet)
                System.out.println(theStringToPrint);

            //Do we output to a log file?
            if (itsLogPrintWriter != null)
                itsLogPrintWriter.println(theStringToPrint);
        }
        catch (Exception e)
        {
        }
    }

    /**
     * Outputs the given string to the logs, including a prefix date and
     * trailing newline, only if the passed in debug level is less than or equal
     * to its debug level.  Use {@link #print(String, int)} if you don't want a
     * prefix date and automatic newline.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inString      the <code>String</code> to output
     * @param inDebugLevel  the debug level of this output (normal is
     *                      <code>0</code>)
     */
    public void println(String inString, int inDebugLevel)
    {
        if (inDebugLevel <= itsDebugLevel)
            println(inString);
    }

    /**
     * Outputs the given <code>Exception</code> to the logs, including a prefix
     * date and trailing newline.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inException  the <code>Exception</code> to output
     */
    public void println(Exception inException)
    {
        try
        {
            //Prepend the date
            Date theCurrentDate = new Date();
            String theStringToPrint = new String("[" +
                                                 getDateString() +
                                                 "] EXCEPTION: " +
                                                 inException.toString());
            //Do we output to System.out?
            if (!itsIsQuiet)
                System.out.println(theStringToPrint);

            //Do we output to a log file?
            if (itsLogPrintWriter != null)
                itsLogPrintWriter.println(theStringToPrint);
        }
        catch (Exception e)
        {
        }
    }

    /**
     * Outputs the given <code>Exception</code> to the logs, including a prefix
     * date and trailing newline, only if the passed in debug level is less than
     * or equal to its debug level.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inException   the <code>Exception</code> to output
     * @param inDebugLevel  the debug level of this output (normal is
     *                      <code>0</code>)
     */
    public void println(Exception inException, int inDebugLevel)
    {
        if (inDebugLevel <= itsDebugLevel)
            println(inException);
    }
}
