/*
JRev: A JAVA MP3 Source Streamer
Copyright (C) 2000  Thomas Bohmbach, Jr.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package org.jrev;

/**
 * A class to handle output to a {@link Logger}.  Classes can extend this
 * class or use an instance of it for delegation.  It provides the same
 * <code>print</code> and <code>println</code> methods that {@link Logger} does
 * as it is a wrapper class.
 *
 * @author      pitboss
 * @version     1.0, 20000728
 * @since       1.0
 * @see         Logger
 */
public class LoggerClient
{
    private Logger itsLogger = null;

    /**
     * Creates a simple instance of a <code>LoggerClient</code> that doesn't log
     * anywhere (because it doesn't have a <code>Logger</code>).
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    public LoggerClient()
    {

    }

    /**
     * Creates an instance of a <code>Logger</code> that has a
     * <code>Logger</code> associated with it.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inLogger  the <code>Logger</code> to output to
     */
    public LoggerClient(Logger inLogger)
    {
        setLogger(inLogger);
    }

    /**
     * Gets the <code>Logger</code> that will be used for logging.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the <code>Logger</code> associated with this
     *              <code>LoggerClient</code>
     */
    public Logger getLogger()
    {
        return itsLogger;
    }

    /**
     * Sets the <code>Logger</code> that will be used for logging.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inLogger  the new <code>Logger</code> to log output to
     */
    public void setLogger(Logger inLogger)
    {
        itsLogger = inLogger;
    }

    /**
     * Stops all logging.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    public void stopLogging()
    {
        itsLogger = null;
    }

    /**
     * Returns whether this <code>LoggerClient</code> is actively logging.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      <code>true</code> if it is actively logging,
     *              <code>false</code> otherwise
     */
    public boolean isLogging()
    {
        boolean theIsLogging = false;

        if (itsLogger != null)
            theIsLogging = true;

        return theIsLogging;
    }

    /**
     * Outputs the given string to its <code>Logger</code>.  No prefix date or
     * newline will be output.  Use {@link #println(String)} if you need a
     * prefix date or automatic newline.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inString  the <code>String</code> to output
     */
    protected void print(String inString)
    {
        if (itsLogger != null)
            itsLogger.print(inString);
    }

    /**
     * Outputs the given string to its <code>Logger</code> only if the passed in
     * debug level is less than or equal to its debug level.  No prefix date or
     * newline will be output.  Use {@link #println(String, int)} if you need a
     * prefix date or automatic newline.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inString      the <code>String</code> to output
     * @param inDebugLevel  the debug level of this output (normal is
     *                      <code>0</code>)
     */
    protected void print(String inString, int inDebugLevel)
    {
        if (itsLogger != null)
            itsLogger.print(inString, inDebugLevel);
    }

    /**
     * Outputs a newline to the its <code>Logger</code>.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    protected void println()
    {
        if (itsLogger != null)
            itsLogger.println();
    }

    /**
     * Outputs the given string to its <code>Logger</code>, including a prefix
     * date and trailing newline.  Use {@link #print(String)} if you don't want
     * a prefix date and automatic newline.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inString  the <code>String</code> to output
     */
    protected void println(String inString)
    {
        if (itsLogger != null)
            itsLogger.println(inString);
    }

    /**
     * Outputs the given string to its <code>Logger</code>, including a prefix
     * date and trailing newline, only if the passed in debug level is less than
     * or equal to its debug level.  Use {@link #print(String, int)} if you
     * don't want a prefix date and automatic newline.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inString      the <code>String</code> to output
     * @param inDebugLevel  the debug level of this output (normal is
     *                      <code>0</code>)
     */
    protected void println(String inString, int inDebugLevel)
    {
        if (itsLogger != null)
            itsLogger.println(inString, inDebugLevel);
    }

    /**
     * Outputs the given <code>Exception</code> to its <code>Logger</code>,
     * including a prefix date and trailing newline.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inException  the <code>Exception</code> to output
     */
    protected void println(Exception inException)
    {
        if (itsLogger != null)
            itsLogger.println(inException);
    }

    /**
     * Outputs the given <code>Exception</code> to its <code>Logger</code>, including a prefix
     * date and trailing newline, only if the passed in debug level is less than
     * or equal to its debug level.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inException   the <code>Exception</code> to output
     * @param inDebugLevel  the debug level of this output (normal is
     *                      <code>0</code>)
     */
    protected void println(Exception inException, int inDebugLevel)
    {
        if (itsLogger != null)
            itsLogger.println(inException, inDebugLevel);
    }
}