/*
JRev: A JAVA MP3 Source Streamer
Copyright (C) 2000  Thomas Bohmbach, Jr.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package org.jrev;

import java.io.File;
import java.io.FileInputStream;
import de.vdheide.mp3.MP3File;
import de.vdheide.mp3.TagContent;

/**
 * An implementation of {@link StreamSource} to use MP3 files as a source.
 *
 * @author      pitboss
 * @version     1.0, 20000728
 * @since       1.0
 * @see         StreamSource
 */
public class MP3FileSource implements StreamSource
{
    private String itsFilename;
    private File itsFile;
    private boolean itsHasReadID3Tags = false;
    private FileInputStream itsFileInputStream = null;
    private long itsFileSizeInBytes = 0;
    private long itsTotalNumBytesRead = 0;
    private float itsPercentComplete = 0.0F;
    private String itsTitle = null;
    private long itsTimeInSeconds = -1;
    private int itsBitrate = -1;
    private int itsSamplerate = -1;
    private String itsAlbum = null;
    private String itsArtist = null;
    private String itsTrack = null;

    /**
     * Creates an <code>MP3FileSource</code> from a file name.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inFilename    the file name of the MP3 file to use as a source
     */
    public MP3FileSource(String inFilename)
    {
        itsFilename = inFilename;
        itsFile = new File(itsFilename);
        itsFileSizeInBytes = itsFile.length();
    }

    /**
     * Cleans up this <code>MP3FileSource</code> upon garbage collection.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    protected void finalize()
    {
        close();
    }

    /**
     * Opens its MP3 file and prepares for reading.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      <code>true</code> if the MP3 is ready to be read from,
     *              <code>false</code> otherwise
     */
    public boolean open()
    {
        boolean theIsOpen = false;

        if (itsFile != null)
        {
            try
            {
                itsFileInputStream = new FileInputStream(itsFile);
                itsTotalNumBytesRead = 0;
                itsPercentComplete = 0.0F;
                theIsOpen = true;
            }
            catch (Exception e)
            {
            }
        }

        return theIsOpen;
    }

    /**
     * Reads byte data into the given buffer from its MP3 file.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inByteArray[]     the byte-buffer to fill
     *
     * @return      the number of bytes read into the buffer, <code>-1</code>
     *              if there's nothing left to read from its MP3 file
     */
    public int read(byte inByteArray[])
    {
        int theNumBytesRead = -1;

        try
        {
            theNumBytesRead = itsFileInputStream.read(inByteArray);

            //Calculate the new progress
            if (theNumBytesRead != -1)
            {
                itsTotalNumBytesRead += theNumBytesRead;
                itsPercentComplete = (float)itsTotalNumBytesRead / (float)itsFileSizeInBytes;
            }
            else //The file is completely read
            {
                itsPercentComplete = 1.0F;
            }
        }
        catch (Exception e)
        {
            theNumBytesRead = -1;
        }

        return theNumBytesRead;
    }

    /**
     * Closes its MP3 file.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    public void close()
    {
        if (itsFileInputStream != null)
        {
            try
            {
                itsFileInputStream.close();
            }
            catch (Exception e)
            {
            }
        }
    }

    /**
     * Reads the MP3 ID3 tags and sets relevent variables
     *
     * @author      pitboss
     * @version     1.1, 20000806
     */
    private void readID3Tags()
    {
        if (itsFilename != null)
        {
            try
            {
                TagContent theTagContent = null;
                MP3File theMP3File = new MP3File(itsFilename);

                //Get title
                theTagContent = theMP3File.getTitle();
                itsTitle = theTagContent.getTextContent();

                //Get time in seconds
                itsTimeInSeconds = theMP3File.getLength();

                //Get bitrate
                itsBitrate = theMP3File.getBitrate();

                //Get samplerate
                itsSamplerate = theMP3File.getSamplerate();

                //Get artist
                theTagContent = theMP3File.getArtist();
                itsArtist = theTagContent.getTextContent();

                //Get album
                theTagContent = theMP3File.getAlbum();
                itsAlbum = theTagContent.getTextContent();

                //Get track
                theTagContent = theMP3File.getTrack();
                itsTrack = theTagContent.getTextContent();
            }
            catch (Exception e)
            {
                System.out.println(e);
            }
        }
    }

    /**
     * Returns the percent of its MP3 file that has already been read.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      a <code>float</code> between <code>0.0</code> and
     *              <code>1.0</code> signifying what percent of this MP3 file
     *              has already been read
     */
    public float percentComplete()
    {
        return itsPercentComplete;
    }

    /**
     * Returns the length of its MP3 file in bytes.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the length of its MP3 file in bytes
     */
    public long lengthInBytes()
    {
        return itsFileSizeInBytes;
    }

    /**
     * Returns the length of its MP3 file in seconds.  This method is TBD.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the length of its MP3 file in seconds
     */
    public long lengthInSeconds()
    {
        if (!itsHasReadID3Tags)
            readID3Tags();

        return itsTimeInSeconds;
    }

    /**
     * Returns the file name of its MP3 file.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the file name of its MP3 file
     */
    public String name()
    {
        return itsFilename;
    }

    /**
     * Returns the title of its MP3 file.  This method currently just strips
     * the path information from the file name and returns it.  TBD, read the
     * title from the ID3 tags.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the title of its MP3 file
     */
    public String title()
    {
        if (!itsHasReadID3Tags)
            readID3Tags();

        return itsTitle;
    }

    /**
     * Returns the bitrate of its MP3 file.
     *
     * @author      pitboss
     * @version     1.1, 20000806
     *
     * @return      the bitrate of its MP3 file
     */
    public int bitrate()
    {
        if (!itsHasReadID3Tags)
            readID3Tags();

        return itsBitrate;
    }

    /**
     * Returns the samplerate of its MP3 file.
     *
     * @author      pitboss
     * @version     1.1, 20000806
     *
     * @return      the samplerate of its MP3 file
     */
    public int samplerate()
    {
        if (!itsHasReadID3Tags)
            readID3Tags();

        return itsSamplerate;
    }

    /**
     * Returns the album that its MP3 file is from
     *
     * @author      pitboss
     * @version     1.1, 20000807
     *
     * @return      the album that its MP3 file is from
     */
    public String album()
    {
        if (!itsHasReadID3Tags)
            readID3Tags();

        return itsAlbum;
    }

    /**
     * Returns the artist of its MP3 file
     *
     * @author      pitboss
     * @version     1.1, 20000807
     *
     * @return      the artist of its MP3 file
     */
    public String artist()
    {
        if (!itsHasReadID3Tags)
            readID3Tags();

        return itsArtist;
    }

    /**
     * Returns the track of its MP3 file
     *
     * @author      pitboss
     * @version     1.1, 20000807
     *
     * @return      the track of its MP3 file
     */
    public String track()
    {
        if (!itsHasReadID3Tags)
            readID3Tags();

        return itsTrack;
    }
}

