/*
JRev: A JAVA MP3 Source Streamer
Copyright (C) 2000  Thomas Bohmbach, Jr.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package org.jrev;

/**
 * An interface that classes implement to provide
 * streaming data (that is, bytes of information) to the {@link JRev}
 * engine.  These classes are the ones that actually do the "reading" of the
 * information.
 *
 * @author      pitboss
 * @version     1.0, 20000728
 * @since       1.0
 * @see         JRev
 */
public interface StreamSource
{
    /**
     * Opens this source.  Even if an implementing class doesn't have a concept
     * of being "open" (like a <code>File</code>) it still needs to implement
     * this method to indicate whether the <code>JRev</code> engine can call
     * {@link #read(byte[])}.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      <code>true</code> if the source is ready to be read from,
     *              <code>false</code> otherwise
     */
    public boolean open();

    /**
     * Reads byte data into the given buffer from this source.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inByteArray[]     the byte-buffer to fill
     *
     * @return      the number of bytes read into the buffer, <code>-1</code>
     *              if there's nothing left to read from this source
     */
    public int read(byte inByteArray[]);

    /**
     * Closes this source.  This method is called when a source is no longer
     * needed by the <code>JRev</code> engine.  If an implementing class
     * doesn't have the concept of being "closed" (like a <code>File</code>),
     * it doesn't need to do anything meaningful with this method.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    public void close();

    /**
     * Returns the percent of this source that has already been read.  If an
     * implementing class doesn't have a way to calculate this (that is, if the
     * total number of bytes isn't known) they should return <code>0.0</code>.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      a <code>float</code> between <code>0.0</code> and
     *              <code>1.0</code> signifying what percent of this source
     *              has already been read
     */
    public float percentComplete();

    /**
     * Returns the name of this source.  For many sources, this will actually
     * be the complete filename or some other method of completely describing a
     * source.  Thus, this is probably not what you want to use for output.
     * If you need the actual title of a particular source, use
     * {@link #title()}.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the name (long description) of this source
     */
    public String name();

    /**
     * Returns the title of this source.  This should be the simplest way to
     * label a source (for example, it should not be the complete filename).
     * This method is provided because many sources will have two (or more) ways
     * to describe them.  In general, use <code>title()</code> when you want to
     * output the title of a source (since, in general it will be shorter).  By
     * default, the <code>JRev</code> engine will print the
     * <code>title()</code> of each source it plays.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the title (short description) of this source
     */
    public String title();

    /**
     * Returns the bitrate of this source.
     *
     * @author      pitboss
     * @version     1.1, 20000811
     *
     * @return      the bitrate of this source
     */
    public int bitrate();
}
