/*
JRev: A JAVA MP3 Source Streamer
Copyright (C) 2000  Thomas Bohmbach, Jr.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package org.jrev;

import java.net.ServerSocket;
import java.net.Socket;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.InterruptedIOException;

/**
 * A class to provide a telnet interface to the {@link JRev} engine and
 * a {@link FilePlaylist} (which must be the {@link Playlist} for the
 * <code>JRev</code> engine).  It listens on a socket for connections and
 * validates users with a password.  Currently, it allows the user to skip the
 * current source, randomize the playlist, rehash the playlist, and shutdown the
 * <code>JRev</code> instance.
 *
 * @author      pitboss
 * @version     1.0, 20000728
 * @since       1.0
 * @see         JRev
 * @see         Playlist
 * @see         FilePlaylist
 */
public class TelnetRemoteInterface extends Thread
{
    private int PASSWORD_TRIES = 3;
    private String HELP_COMMAND = "help";
    private String INFO_COMMAND = "info";
    private String SKIP_COMMAND = "skip";
    private String RANDOMIZE_COMMAND = "randomize";
    private String REHASH_COMMAND = "rehash";
    private String RESTART_COMMAND = "restart";
    private String SHUTDOWN_COMMAND = "shutdown";
    private String EXIT_COMMAND = "exit";

    private JRev itsStreamer = null;
    private FilePlaylist itsPlaylist = null;
    private int itsPort = 8002;
    private String itsPassword = "";
    private LoggerClient itsLoggerClient = null;
    private ServerSocket itsServerSocket = null;
    private BufferedReader itsBufferedReader = null;
    private PrintWriter itsPrintWriter = null;
    private Thread itsThread = null;
    private boolean itsKeepListening = true;

    /**
     * Creates an instance of a <code>TelnetRemoteInterface</code> that doesn't
     * log.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inPort        the port to listen for connections on
     * @param inPassword    the password to validate connections with
     * @param inStreamer    the <code>JRev</code> instance that this
     *                      <code>TelnetRemoteInterface</code> will control
     * @param inPlaylist    the <code>FilePlaylist</code> instance that this
     *                      <code>TelnetRemoteInterface</code> will control
     */
    public TelnetRemoteInterface(int inPort,
                                 String inPassword,
                                 JRev inStreamer,
                                 FilePlaylist inPlaylist)
    {
        itsPort = inPort;
        itsPassword = inPassword;
        itsStreamer = inStreamer;
        itsPlaylist = inPlaylist;
        itsLoggerClient = new LoggerClient();

        init();
    }

    /**
     * Creates an instance of a <code>TelnetRemoteInterface</code> that logs to
     * a <code>Logger</code>.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inPort        the port to listen for connections on
     * @param inPassword    the password to validate connections with
     * @param inStreamer    the <code>JRev</code> instance that this
     *                      <code>TelnetRemoteInterface</code> will control
     * @param inPlaylist    the <code>FilePlaylist</code> instance that this
     *                      <code>TelnetRemoteInterface</code> will control
     * @param inLogger      the <code>Logger</code> to send output to
     */
    public TelnetRemoteInterface(int inPort,
                                 String inPassword,
                                 JRev inStreamer,
                                 FilePlaylist inPlaylist,
                                 Logger inLogger)
    {
        itsPort = inPort;
        itsPassword = inPassword;
        itsStreamer = inStreamer;
        itsPlaylist = inPlaylist;

        itsLoggerClient = new LoggerClient(inLogger);

        init();
    }

    /**
     * A generic initialization method for all constructors.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    private void init()
    {
        try
        {
            itsLoggerClient.println("Starting TRI on port " + itsPort + "...");
            itsServerSocket = new ServerSocket(itsPort);

            if (itsServerSocket != null)
            {
                itsServerSocket.setSoTimeout(1000);
                itsThread = new Thread(this);
                itsThread.start();
            }
        }
        catch (Exception e)
        {
            itsLoggerClient.println(e);
        }
    }

    /**
     * Cleans up this <code>TelnetRemoteInterface</code> upon garbage
     * collection.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    protected void finalize()
    {
        try
        {
            if (itsServerSocket != null)
                itsServerSocket.close();

            if (itsPrintWriter != null)
                itsPrintWriter.close();

            if (itsBufferedReader != null)
                itsBufferedReader.close();
        }
        catch (Exception e)
        {
            itsLoggerClient.println(e);
        }

        itsServerSocket = null;
        itsPrintWriter = null;
        itsBufferedReader = null;
    }

    /**
     * Outputs the given string to the remote user and the logs.
     * No prefix date or newline will be output to the remote user, but will be
     * added to the log.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inString  the <code>String</code> to output
     */
    private void message(String inString)
    {
        try
        {
            itsLoggerClient.println(inString);  //To the logger
            itsPrintWriter.println(inString + "\r");
        }
        catch (Exception e)
        {
            itsLoggerClient.println(e);
        }
    }

    /**
     * Outputs the command prompt to the remote user and the logs.  No prefix
     * date or newline will be output.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    private void commandPrompt()
    {
        try
        {
            itsLoggerClient.print("-> ");   //To the logger
            itsPrintWriter.print("-> ");
            itsPrintWriter.flush();
        }
        catch (Exception e)
        {
            itsLoggerClient.println(e);
        }
    }

    /**
     * Returns whether its <code>JRev</code> instance is running.
     *
     * @author      pitboss
     * @version     1.1, 20000813
     *
     * @return      <code>true</code> if its <code>JRev</code> instance is
     *              running, <code>false</code> otherwise
     */

    private boolean isJRevRunning()
    {
        boolean theIsRunning = false;

        if (itsStreamer != null)
            theIsRunning = itsStreamer.isRunning();

        return theIsRunning;
    }

    /**
     * Gets the name of its <code>JRev</code> stream.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the name of the stream
     */
    private String getStreamName()
    {
        String theStreamName = "JRev";

        if (itsStreamer != null)
            theStreamName = itsStreamer.name();

        return theStreamName;
    }

    /**
     * Gets the title of its <code>JRev</code> current source.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @return      the title of the current source
     */
    private String getCurrentTitle()
    {
        String theTitle = "current source";

        if (itsStreamer != null)
        {
            StreamSource theSource = itsStreamer.getSource();

            if (theSource != null)
                theTitle = theSource.title();

            if (theTitle == null)
                theTitle = "current source";
        }

        return theTitle;
    }

    /**
     * A utility method for turning seconds into MM:SS
     *
     * @author      pitboss
     * @version     1.1, 20000807
     *
     * @param inSeconds     the number of seconds to convert
     *
     * @return              A string in the form of MM:SS
     */
    private String secondsToTimeString(long inSeconds)
    {
        long theMinutes = inSeconds / 60;
        long theSeconds = inSeconds - (theMinutes * 60);
        StringBuffer theMMSSStringBuffer = new StringBuffer();

        theMMSSStringBuffer.append(theMinutes);

        theMMSSStringBuffer.append(":");

        if (theSeconds < 10)
            theMMSSStringBuffer.append("0");
        theMMSSStringBuffer.append(theSeconds);

        return (theMMSSStringBuffer.toString());
    }

    /**
     * Gets the info string of its <code>JRev</code> current source.
     *
     * @author      pitboss
     * @version     1.1, 20000811
     *
     * @return      the info string of the current source
     */
    private String getCurrentInfo()
    {
        String theInfoString = "No information available.";

        if (itsStreamer != null)
        {
            MP3FileSource theSource = (MP3FileSource) itsStreamer.getSource();

            if (theSource != null)
            {
                StringBuffer theStringBuffer = new StringBuffer();
                String theString = null;

                //Append the filename
                theString = theSource.name();
                if (theString != null)
                    theStringBuffer.append("Filename:   " + theString + "\n\r");

                //Append the title
                theString = theSource.title();
                if (theString != null)
                    theStringBuffer.append("Title:      " + theString + "\n\r");

                //Append the artist
                theString = theSource.artist();
                if (theString != null)
                    theStringBuffer.append("Artist:     " + theString + "\n\r");

                //Append the album
                theString = theSource.album();
                if (theString != null)
                    theStringBuffer.append("Album:      " + theString + "\n\r");

                //Append the time
                long theTimeInSeconds = theSource.lengthInSeconds();
                if (theTimeInSeconds != 0)
                {
                    theString = secondsToTimeString(theTimeInSeconds);
                    theStringBuffer.append("Total Time: " + theString + "\n\r");
                }

                float thePercentComplete = theSource.percentComplete();
                long theTimeRemaining = theTimeInSeconds -
                    (long)(theTimeInSeconds * thePercentComplete);
                if (theTimeRemaining > 0)
                {
                    theString = secondsToTimeString(theTimeRemaining);
                    theStringBuffer.append("Remaining:  " + theString + "\n\r");
                }

                theInfoString = theStringBuffer.toString();
            }
        }

        return theInfoString;
    }

    /**
     * Restarts the <code>JRev</code> instance that its responsible for.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    private void restartStreamer()
    {
        message("Restarting " + getStreamName() + "...");
        if (itsStreamer != null)
            itsStreamer.restart();

        message("");
    }

    /**
     * Shuts down the <code>JRev</code> and <code>FilePlaylist</code>
     * instances that its responsible for.  Note that this will kill this
     * <code>TelnetRemoteInterface</code> as well.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    private void shutdownStreamer()
    {
        if (itsStreamer != null)
        {
            itsStreamer.shutdown();
            itsStreamer = null;
        }

        if (itsPlaylist != null)
        {
            itsPlaylist.close();
            itsPlaylist = null;
        }

        itsKeepListening = false;
    }

    /**
     * Gets the info on the current source.
     *
     * @author      pitboss
     * @version     1.1, 20000811
     */
    private void sourceInfo()
    {
        if (itsStreamer != null)
        {
            String theInfo = getCurrentInfo();

            message(theInfo);
        }
    }

    /**
     * Skips the current source.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    private void skipSource()
    {
        if (itsStreamer != null)
        {
            String theTitle = getCurrentTitle();

            message("Skipping " + theTitle + "...");

            itsStreamer.skip();

            try
            {
                sleep(1000);
            }
            catch (InterruptedException e)
            {
            }

            theTitle = getCurrentTitle();

            message("Now playing " + theTitle + ".");
            message("");
        }
    }

    /**
     * Re-reads the playlist.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    private void rehashPlaylist()
    {
        if (itsPlaylist != null)
        {
            message("Rehashing playlist...");
            itsPlaylist.rehash();
            message("");
        }
    }

    /**
     * Re-randomizes the playlist.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    private void randomizePlaylist()
    {
        if (itsPlaylist != null)
        {
            message("Randomizing playlist...");
            itsPlaylist.randomize();
            message("");
        }
    }

    /**
     * The main <code>Thread</code> method.  Listens on the port for a
     * connection and routes commands to a handler.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    public void run()
    {
        Socket theSocket = null;

        try
        {
            itsLoggerClient.println("Waiting for a remote connection...");

            while ((itsServerSocket != null) &&
                   isJRevRunning() &&
                   itsKeepListening)
            {
                theSocket = null;

                try
                {
                    theSocket = itsServerSocket.accept();

                    handleConnection(theSocket);
                }
                catch (InterruptedIOException iioe)
                {
                    //Do nothing...
                }
                catch (Exception e)
                {
                    itsLoggerClient.println(e);
                }
            }

            //Did JRev stop running?
            if (! isJRevRunning())
            {
                //Do we need to close the playlist?
                if (itsPlaylist != null)
                    itsPlaylist.close();
            }
        }
        catch (Exception re)
        {
            itsLoggerClient.println(re);
        }
        finally
        {
            try
            {
                if (itsServerSocket != null)
                {
                    itsLoggerClient.println("Stopping TRI on port " +
                                            itsPort +
                                            ".");
                    itsServerSocket.close();
                }
            }
            catch (Exception fe)
            {
                itsLoggerClient.println(fe);
            }

            itsServerSocket = null;
        }
    }

    /**
     * Handles a remote user connection.  Initializes connection with user and
     * verifies password.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inSocket      the <code>Socket</code> to communicate on
     */
    private void handleConnection(Socket inSocket)
    {
        String thePassword = null;
        Socket theSocket = inSocket;
        boolean theIsStayConnected = true;

        if (inSocket != null)
        {
            try
            {
                itsBufferedReader = new BufferedReader
                    (new InputStreamReader(theSocket.getInputStream()));
                itsPrintWriter = new PrintWriter(theSocket.getOutputStream(),
                                                 true);

                String theStreamName = getStreamName();

                message("Welcome to the JRev Telnet Remote Interface.");
                message("This TRI controls " + theStreamName + ".");

                for (int i = 0; i < PASSWORD_TRIES; i++)
                {
                    message("Please enter the password:");

                    thePassword = readCommand();

                    if (itsPassword.equals(thePassword))
                    {
                        //Don't loop again
                        i = PASSWORD_TRIES;
                        message("Login successful.  Ready for commands.");
                        message("(Type 'help' for a list of commands).");
                        message("");

                        while (theIsStayConnected && isJRevRunning())
                        {
                            commandPrompt();
                            theIsStayConnected = handleCommand(readCommand());
                        }
                    }
                    else
                    {
                        message("Incorrect password.");
                    }
                }

                message("Goodbye.");
            }
            catch (Exception e)
            {
                itsLoggerClient.println(e);
            }
            finally
            {
                try
                {
                    if (itsPrintWriter != null)
                        itsPrintWriter.close();

                    if (itsBufferedReader != null)
                        itsBufferedReader.close();

                    if (theSocket != null)
                        theSocket.close();
                }
                catch (Exception fe)
                {
                    itsLoggerClient.println(fe);
                }

                itsPrintWriter = null;
                itsBufferedReader = null;
            }
        }
    }

    /**
     * Reads a command from the remote user.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     */
    private String readCommand()
    {
        String theInputString = null;

        if (itsBufferedReader != null)
        {
            try
            {
                //Wait for there to be input before reading (and blocking)
                while ((! itsBufferedReader.ready()) &&
                       isJRevRunning())
                {
                    sleep(1000);
                }

                //Are we still running?
                if (isJRevRunning())
                {
                    theInputString = itsBufferedReader.readLine();

                    if (theInputString != null)
                    {
                        theInputString = theInputString.trim();
                        itsLoggerClient.print(theInputString + "\n"); //To the logger
                    }
                }
            }
            catch (Exception e)
            {
                itsLoggerClient.println(e);
            }
        }

        return theInputString;
    }

    /**
     * Handles commands from the remote user by routing them to the appropriate
     * command handlers.
     *
     * @author      pitboss
     * @version     1.0, 20000728
     *
     * @param inCommand     the command to execute
     */
    private boolean handleCommand(String inCommand)
    {
        boolean theIsStayConnected = true;

        if (inCommand != null)
        {
            //Print the menu of commands
            if (inCommand.equalsIgnoreCase(HELP_COMMAND))
            {
                message("Command   : Explanation");
                message("-------------------------------------------");
                message("help      : Print this list of commands.");
                message("info      : Get source information.");
                message("skip      : Skip to the next source.");
                message("randomize : Randomize the playlist.");
                message("rehash    : Re-read the playlist.");
                message("restart   : Restart this JRev instance.");
                message("shutdown  : Stop this JRev instance.");
                message("exit      : Exit the remote interface.");
                message("");
            }
            else if (inCommand.equalsIgnoreCase(INFO_COMMAND))
            {
                sourceInfo();
            }
            else if (inCommand.equalsIgnoreCase(SKIP_COMMAND))
            {
                skipSource();
            }
            else if (inCommand.equalsIgnoreCase(RANDOMIZE_COMMAND))
            {
                randomizePlaylist();
            }
            else if (inCommand.equalsIgnoreCase(REHASH_COMMAND))
            {
                rehashPlaylist();
            }
            else if (inCommand.equalsIgnoreCase(RESTART_COMMAND))
            {
                restartStreamer();
            }
            else if (inCommand.equalsIgnoreCase(SHUTDOWN_COMMAND))
            {
                theIsStayConnected = false;
                shutdownStreamer();
            }
            else if (inCommand.equalsIgnoreCase(EXIT_COMMAND))
            {
                theIsStayConnected = false;
            }
            else
            {
                message("Unrecognized command: " + inCommand);
                message("Please use 'help' for a list of valid commands.");
                message("");
            }
        }

        return theIsStayConnected;
    }
}
